#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import functools
import inspect
import logging
import re
import warnings
from collections.abc import Generator
from hashlib import sha1
from typing import cast

import pandas as pd

__all__ = [
    "get_model",
    "hist_search",
    "clean_feature_name",
    "clean_column_names",
    "deprecated",
    "check_dtypes",
]


def get_model(model_str: str):
    from sklearn import ensemble, linear_model, svm

    from amplo import classification, regression
    from amplo.base import BasePredictor

    model: BasePredictor

    if "RandomForest" in model_str or "Bagging" in model_str:
        model = getattr(ensemble, model_str)()
    elif model_str == "SVC":
        model = svm.SVC(probability=True)  # type: ignore
    elif model_str == "SVR":
        model = svm.SVR()  # type: ignore
    elif "Logistic" in model_str or "Linear" in model_str or "Ridge" in model_str:
        model = getattr(linear_model, model_str)()
    elif "Classifier" in model_str:
        model = getattr(classification, model_str)()
    elif "Regressor" in model_str:
        model = getattr(regression, model_str)()
    else:
        raise ValueError("Model not recognized.")
    return model


def hist_search(array, value):
    """
    Binary search that finds the index in order to fulfill
    ``array[index] <= value < array[index + 1]``

    Parameters
    ----------
    array : array of float
    value : float

    Returns
    -------
    int
        Bin index of the value
    """

    # Return -1 when no bin exists
    if value < array[0] or value >= array[-1]:
        logging.debug(
            f"No bin (index) found for value {value}. "
            f"Array(Min: {array[0]}, "
            "Max: {array[-1]})"
        )
        return -1

    # Initialize min and max bin index
    low = 0
    high = len(array) - 1

    # Bin search
    countdown = 30
    while countdown > 0:
        # Count down
        countdown -= 1

        # Set middle bin index
        middle = low + (high - low) // 2

        if low == middle == high - 1:  # stop criterion
            return middle

        if value < array[middle]:  # array[low] <= value < array[middle]
            high = middle
        elif value >= array[middle]:  # array[middle] <= value < array[high]
            low = middle

    warnings.warn("Operation took too long. Returning -1 (no match).", RuntimeWarning)
    return -1


def clean_feature_name(feature_name: str | int, *, hash_size: int = 0) -> str:
    """
    Clean feature names and append "feature_" when it's a digit.

    Parameters
    ----------
    feature_name : str or int
        Feature name to be cleaned.
    hash_size : int, default: 0
        If > 0 it appends a sha1 hashing of size `hash_size` to the input.

    Returns
    -------
    cleaned_feature_name : str
    """

    # Handle feature_name
    if isinstance(feature_name, int) or str(feature_name).isdigit():
        f_name = f"feature_{feature_name}"  # append "feature_" when it's an integer
    else:
        f_name = str(feature_name)

    # Append hash
    if hash_size > 0:
        hash_string = sha1(f_name.encode()).hexdigest()
        f_name = f"{f_name}_{hash_string[:hash_size]}"

    # Remove non-numeric and non-alphabetic characters.
    # Assert single underscores and remove underscores in prefix and suffix.
    return re.sub("[^a-z0-9]+", "_", f_name.lower()).strip("_")


def clean_column_names(
    data: pd.DataFrame, *, hash_duplicates: bool = True
) -> tuple[pd.DataFrame, dict[str, str]]:
    """
    Cleans column names while asserting their uniqueness.

    Parameters
    ----------
    data : pd.DataFrame
        Data to be cleaned.
    hash_duplicates : bool
        Whether to hash duplicated renamings.

    Returns
    -------
    pd.DataFrame
        Same data but with cleaned column names.
    dict of {str : str}
        Dictionary which indicates the renaming.

    Raises
    ------
    ValueError
        When duplicated column names are produced.
    """

    # Make first renaming attempt. May create duplicated names.
    renaming = pd.Series({old: clean_feature_name(old) for old in data.columns})

    if hash_duplicates:
        # Check if duplicates in the renaming exist and hash those
        duplicated = renaming.duplicated(keep=False)  # mark all duplicates as True
        if duplicated.any():
            renaming[duplicated] = (
                pd.Series(renaming[duplicated].index)
                .apply(clean_feature_name, hash_size=4)
                .to_list()
            )

    # Raise error if (still) contains duplicates
    duplicated = renaming.duplicated()
    if duplicated.any():
        raise ValueError(
            f"Duplicate column names exist that causes ambiguity: "
            f"{renaming[duplicated].index.to_list()}"
        )

    # Convert to dictionary and apply
    renaming = renaming.to_dict()
    renaming = cast(dict[str, str], renaming)

    return data.rename(columns=renaming), renaming


def deprecated(reason):
    # This decorator is a copy-pase from:
    # https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """
    string_types = (type(b""), type(""))

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


def check_dtypes(*dtype_tuples):
    """
    Checks all dtypes of given list.

    Parameters
    ----------
    *dtype_tuples : Any
        Tuples of (name, parameter, allowed types) to be checked.
        When checking only one parameter, the wrapping in a tuple can be omitted.

    Returns
    -------
    None

    Examples
    --------
    Check a single parameter:
    >>> check_dtypes(("var1", 123, int))  # tuple
    >>> check_dtypes("var1", 123, int)  # without tuple

    Check multiple:
    >>> check_dtypes(("var1", 123, int), ("var2", 1.0, (int, float)))  # tuples
    >>> check_dtypes(("var", var, str) for var in ["a", "b"])  # list or generator

    Raises
    ------
    ValueError
        If any given constraint is not fulfilled.
    """
    # Allow single dtype check without wrapping in a tuple
    if isinstance(dtype_tuples[0], str):
        if len(dtype_tuples) != 3:
            raise TypeError("Invalid arguments for `check_dtypes` function.")
        dtype_tuples = (dtype_tuples,)

    # Allow single list or generator object
    if isinstance(dtype_tuples[0], (list, Generator)):
        if len(dtype_tuples) != 1:
            raise TypeError("Invalid arguments for `check_dtypes` function.")
        dtype_tuples = dtype_tuples[0]

    # Check dtypes
    for name, value, typ in dtype_tuples:
        if not isinstance(value, typ):
            msg = f"Invalid dtype for argument `{name}`: {type(value).__name__}"
            raise TypeError(msg)



#  Copyright (c) 2022 by Amplo.
from __future__ import annotations

import faiss
import numpy as np
import pandas as pd

from amplo.base import LoggingMixin


class IntervalAnalyser(LoggingMixin):

    noise = -1

    def __init__(
        self,
        target: str,
        norm: str = "euclidean",  # TODO: implement functionality
        min_length: int = 1000,
        n_neighbors: int | None = None,
        n_trees: int = 10,
        verbose: int = 1,
    ):
        """
        Interval Analyser for Log file classification. Has two purposes:
        - Remove healthy data in longer, faulty logs
        - Remove redundant data in large datasets

        Uses Facebook's FAISS for K-Nearest Neighbors approximation.

        NOTE: This should no longer be used as a standalone module, and should be used
        with the Pipeline. It can still be used as as a standalone module, but it needs
        to take normalized, multi-indexed data.

        Parameters
        ----------
        target : str
            Target column name, must be a folder in the parent folder!
        norm : str
            Optimization metric for K-Nearest Neighbors
            NOTE: This option has no effect, yet!
        min_length : int
            Minimum length to cut off, everything shorter is left untouched
        n_neighbors : int, optional
            Quantity of neighbors, default to 3 * log length
        n_trees : int
            Quantity of trees
        """
        super().__init__(verbose=verbose)
        # Test
        self.available_norms = ["euclidean", "manhattan", "angular", "hamming", "dot"]
        if norm not in self.available_norms:
            raise ValueError(f"Unknown norm, pick from {self.available_norms}")

        # Parameters
        self.target = target
        self.min_length = min_length
        self.norm = norm
        self.n_trees = n_trees
        self.n_neighbors = n_neighbors
        self.verbose = verbose

        # Initializers
        self._engine = None
        self.distributions_ = None
        self.noise_indices_ = None
        self.n_samples_ = None
        self.n_columns_ = None
        self.n_files_ = None
        self.n_folders_ = None

        # Flags
        self.is_fitted_ = False

    def fit_transform(self, data: pd.DataFrame):
        """
        Function that runs the K-Nearest Neighbors and returns a dataframe with the
        sensitivities.

        Parameters
        ----------
        data : pd.DataFrame
            Multi-indexed dataset containing feature (and target) columns
            or path to folder that is correctly structured (see Notes)
        labels : pd.Series, optional
            Multi-indexed dataset containing target columns.

        Returns
        -------
        pd.DataFrame
            Data with noise filtered
        """
        # Validate data
        data = self._validate_data(data)
        self._set_data_attr(data)

        # Set up Annoy Engine (only now that n_keys is known)
        self._build_engine(data)

        # Make distribution
        self.is_fitted_ = True
        return self.transform(data)

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Filters out noise for the provided data frame.
        """
        if not self.is_fitted_:
            raise ValueError("Object not fitted yet.")

        data = self._validate_data(data)
        self._make_distribution(data)
        return self._filter_noise(data)

    def _validate_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Validates the data and sets certain attributes.
        """
        # Validate data
        if not self.target in data:
            raise ValueError(f"Target column {self.target} not in dataframe.")
        if len(data.index.names) != 2:
            raise ValueError(f"Dataframe needs to have a multi-index.")
        if data.max().max() > 10 or data.min().min() < -10:
            raise ValueError("Data needs to be normalized.")

        # Assert that all dtypes are of dtype float32
        #  See: https://github.com/facebookresearch/faiss/issues/461
        for col in data.select_dtypes(exclude=["float32"]).columns:
            data[col] = data[col].astype("float32")

        return data

    def _set_data_attr(self, data: pd.DataFrame):
        """
        Sets data specific class attributes
        """
        self.n_folders_ = data[self.target].nunique()
        self.n_files_ = data.index.get_level_values(0).nunique()
        self.n_samples_, self.n_columns_ = len(data), len(data.keys()) - 1
        if self.n_neighbors is None:
            self.n_neighbors = min(3 * self.n_samples_ // self.n_files_, 5000)

    def _build_engine(self, data: pd.DataFrame):
        """
        Builds the ANNOY engine.

        Parameters
        ----------
        features : pd.DataFrame
            Multi-indexed dataset containing feature columns
        """
        # Create engine
        self.logger.info("Building interval analyser engine.")
        self._engine = faiss.IndexFlatL2(self.n_columns_)

        # Add the data to ANNOY
        data = data.drop(self.target, axis=1)
        self._engine.add(np.ascontiguousarray(data.values))  # type: ignore

    def _make_distribution(self, data: pd.DataFrame):
        """
        Given a build K-Nearest Neighbors, returns the label distribution

        Parameters
        ----------
        features : pd.DataFrame
            Multi-indexed dataset containing feature columns
        labels : pd.Series
            Multi-indexed dataset containing target columns
        """
        assert self._engine and self.n_neighbors
        self.logger.info("Calculating interval within-class distributions.")
        labels = data[self.target]
        data = data.drop(self.target, axis=1)

        # Search nearest neighbors for all samples - has to be iterative for large files
        distribution = []
        for i, row in data.iterrows():
            _, neighbors = self._engine.search(  # type: ignore
                np.ascontiguousarray(np.array(row.values).reshape((1, -1))),
                self.n_neighbors,
            )
            match_mask = labels.iloc[neighbors.reshape(-1)] == labels.loc[i]  # type: ignore
            distribution.append(pd.Series(match_mask).sum() / self.n_neighbors)  # type: ignore

        # Parse into list of lists
        self.distributions_ = pd.Series(distribution, index=labels.index)

    def _filter_noise(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        This function selects samples given the calculated distributions. It only
        removes samples from logs which are longer (> min_length), and only the samples
        with lower in-class neighbors.

        One could come up with a fancier logic, using the total dataset samples, the
        class-balance & sample redundancy.
        """
        assert self.distributions_ is not None
        # Verbose
        self.logger.info("Creating filtered dataset")

        # Initialize
        noise_indices = []

        # Iterate through labels and see if we should remove values
        for log in self.distributions_.index.get_level_values(0):

            # Check distribution and find cut-off
            dist = self.distributions_[log]
            ind_remove_label = [(log, j) for j in np.where(dist < dist.mean())[0]]

            # Extend list to keep track
            noise_indices.extend(ind_remove_label)

            # Verbose
            self.logger.info(
                f"Removing {len(ind_remove_label)} samples from " f"`{log}`"
            )

        # Set noise indices
        self.noise_indices_ = noise_indices
        data = data.drop(self.noise_indices_, axis=0)
        return data



#  Copyright (c) 2022 by Amplo.

import numpy as np
import pandas as pd
import pytest

from amplo.automl.standardization import Standardizer


class TestStandardizer:
    def test_fit_transform(self):
        df = pd.DataFrame(
            {"x1": np.random.normal(10, 1, 1000), "x2": np.random.normal(-10, 15, 1000)}
        )
        s = Standardizer()
        df_standard = s.fit_transform(df)
        assert df_standard.mean().abs().max() < 1e3
        assert df_standard.std().max() < 1.001 and df_standard.std().min() > 0.999
        assert s.cols_ is not None and set(s.cols_) == {"x1", "x2"}
        assert s.means_ is not None and set(s.means_.keys()) == {"x1", "x2"}
        assert s.stds_ is not None and set(s.stds_.keys()) == {"x1", "x2"}

    def test_target_classification(self):
        df = pd.DataFrame(
            {
                "x1": np.random.normal(10, 1, 100),
                "x2": np.random.normal(-10, 15, 100),
                "y": np.random.random_integers(0, 1, 100),
            }
        )
        s = Standardizer(target="y")
        df_standard = s.fit_transform(df)
        assert set(df_standard["y"].unique()) == {0, 1}

    def test_reverse(self):
        df = pd.DataFrame({"x1": np.random.normal(10, 1, 100)})
        s = Standardizer()
        df_standard = s.fit_transform(df)
        reverted = s.reverse(df_standard["x1"].to_numpy().reshape((-1)), "x1")
        assert np.allclose(reverted, df["x1"])



#  Copyright (c) 2022 by Amplo.
from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import KFold, StratifiedKFold

from amplo.automl.grid_search import OptunaGridSearch
from amplo.utils.io import parse_json
from tests import get_all_modeller_models


@pytest.fixture(scope="class", params=["regression", "classification"])
def make_data(request):
    mode = request.param

    if mode == "regression":
        request.cls.k_fold = KFold
        request.cls.objective = "r2"
        x, y = make_regression(n_features=5)  # type: ignore
    elif mode == "classification":
        request.cls.k_fold = StratifiedKFold
        request.cls.objective = "neg_log_loss"
        x, y = make_classification(n_features=5)
    else:
        raise ValueError("Mode is invalid")

    request.cls.mode = mode
    request.cls.data = pd.DataFrame(x)
    request.cls.data["target"] = y
    yield


@pytest.mark.usefixtures("make_data")
class TestGridSearch:
    objective: str
    mode: str
    k_fold: type[KFold] | type[StratifiedKFold]
    data: pd.DataFrame

    def test_each_model(self):
        """
        Given the models which depend on 1) mode and 2) num_samples,
        each model is trained on the given grid search class.

        Thanks to the fact that all grid search models have the same interface,
        this is possible :-)
        """
        models = get_all_modeller_models(mode=self.mode, objective=self.objective)

        for model in models:
            # Grid search
            search = OptunaGridSearch(
                model,
                target="target",
                cv=self.k_fold(n_splits=3),
                verbose=0,
                timeout=10,
                n_trials=2,
                scoring=self.objective,
            )
            results = search.fit(self.data)

            # Tests
            model_name = type(model).__name__
            assert isinstance(results, pd.DataFrame), (
                f"Expected result to be pandas.DataFrame"
                f"but found {type(results)} in {model_name} instead"
            )
            results_keys = [
                "worst_case",
                "score",
                "params",
                "time",
            ]
            assert all(
                key in results.keys() for key in results_keys
            ), f"Keys are missing in {model_name}"
            assert len(results) > 0, f"Results are empty in {model_name}"
            model.set_params(**parse_json(results.iloc[0]["params"]))



#  Copyright (c) 2022 by Amplo.
from __future__ import annotations

from copy import deepcopy
from warnings import warn

import numpy as np
import pandas as pd

from amplo.base import LoggingMixin


class Standardizer(LoggingMixin):
    def __init__(
        self, target: str | None = None, mode: str = "classification", verbose: int = 1
    ):
        super().__init__(verbose=verbose)
        if mode not in ["classification", "regression"]:
            raise ValueError(
                "Invalid mode. Pick from 'classification' and 'regression'"
            )
        self.mode = mode
        self.target = target
        self.cols_: list[str] | None = None
        self.means_: pd.Series | None = None
        self.stds_: pd.Series | None = None
        self.is_fitted_ = False

    def fit_transform(self, data: pd.DataFrame) -> pd.DataFrame:
        self.fit(data)
        self.is_fitted_ = True
        return self.transform(deepcopy(data))

    def fit(self, data: pd.DataFrame):
        # Gather cols
        self.cols_ = data.select_dtypes(include=np.number).columns.tolist()  # type: ignore

        # Remove target if classification
        if self.target and self.target in self.cols_ and self.mode == "classification":
            self.cols_.remove(self.target)

        # Set attrs
        self.means_ = data[self.cols_].mean(axis=0)
        self.stds_ = data[self.cols_].std(axis=0)
        self.stds_[abs(self.stds_) <= 1e-9] = 1

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        if not self.is_fitted_:
            raise ValueError("Object not yet fitted.")
        assert self.cols_ is not None
        if not all(k in data for k in self.cols_):
            warn(f"Missing keys in data: {[k for k in self.cols_ if k not in data]}")

        data[self.cols_] = (data[self.cols_] - self.means_) / self.stds_
        return data

    def reverse(self, data: np.ndarray, column: str | None = None) -> pd.Series:
        """
        Currently only used to reverse predict regression predictions. Hence the natural
        support for np.ndarrays.
        """
        if not self.is_fitted_:
            raise ValueError("Object not yet fitted.")
        assert (
            self.cols_ is not None
            and self.stds_ is not None
            and self.means_ is not None
        )

        # If columns are provided...
        if column:
            if data.ndim != 1:
                raise ValueError(
                    "When column is specified, data should be 1 dimensional."
                )
            data = data * self.stds_[column] + self.means_[column]
            return pd.Series(data)

        # ...and if not
        if len(self.cols_) != len(data[0]):
            raise ValueError("Unequal columns and provided column strings.")
        data = data * self.stds_ + self.means_
        return pd.Series(data)

    def get_settings(self):
        if not self.is_fitted_:
            raise ValueError("Object is not yet fitted.")
        assert self.cols_ and self.means_ is not None and self.stds_ is not None
        return {
            "target": self.target,
            "mode": self.mode,
            "cols_": self.cols_,
            "means_": self.means_.to_json(),
            "stds_": self.stds_.to_json(),
        }

    def load_settings(self, settings: dict):
        self.cols_ = settings.get("cols_", [])
        self.means_ = settings.get("means_")
        self.stds_ = settings.get("stds_")
        self.mode = settings.get("mode")
        self.target = settings.get("target")
        return self



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import json
import os
import time
import warnings
from inspect import signature
from pathlib import Path
from typing import Any
from warnings import warn

import joblib
import numpy as np
import pandas as pd
from shap import TreeExplainer
from sklearn import metrics
from sklearn.model_selection import KFold, StratifiedKFold

import amplo
from amplo.automl.data_processing import DataProcessor
from amplo.automl.feature_processing import FeatureProcessor
from amplo.automl.feature_processing.feature_processor import (
    get_required_columns,
    translate_features,
)
from amplo.automl.grid_search import OptunaGridSearch
from amplo.automl.interval_analysis import IntervalAnalyser
from amplo.automl.modelling import Modeller
from amplo.automl.sequencing import Sequencer
from amplo.automl.standardization import Standardizer
from amplo.base import BasePredictor
from amplo.base.objects import LoggingMixin
from amplo.observation import DataObserver, ModelObserver
from amplo.utils import clean_feature_name, get_model, io, logging
from amplo.validation import ModelValidator

__all__ = ["Pipeline"]

warnings.filterwarnings("ignore", message="lbfgs failed to converge")


class Pipeline(LoggingMixin):
    """
    Automated Machine Learning Pipeline for tabular data.

    The pipeline is designed for predictive maintenance application, failure
    identification, failure prediction, condition monitoring, and more.

    Parameters
    ----------
    # Main parameters
    main_dir : str, default: "Auto_ML/"
        Main directory of pipeline
    target : str, optional
        Column name of the output variable.
    name : str, default: "AutoML"
        Name of the project
    version : int, default: 1
        Pipeline version. Will automatically increment when a version exists.
    mode : {None, "classification", "regression"}, default: None
        Pipeline mode.
    objective : str, optional
        Objective for training.
        Default for classification: "neg_log_loss".
        Default for regression: "mean_square_error".
    verbose : int, default: 1
        Verbosity of logging.
    logging_to_file : bool, default: False
        Whether to write logging to a file
    logging_path : str, default: "AutoML.log"
        Write to logging to given path if ``logs_to_file`` is True.

    # Data processing
    missing_values : {"remove", "interpolate", "mean", "zero"}, default: "zero"
        How to treat missing values.
    outlier_removal : {"clip", "boxplot", "z-score", "none"}, default: "clip"
        How to treat outliers.
    z_score_threshold : int, default: 4
        When ``outlier_removal`` is "z-score", the threshold is adaptable.
    include_output : bool, default: False
        Whether to include output in the training data (sensible only with sequencing).

    # Balancing
    balance : bool, default: False
        Whether to balance data.

    # Feature processing
    extract_features : bool, default: True
        Whether to use the FeatureProcessing module to extract features.
    information_threshold : float, default: 0.999
        Threshold for removing collinear features.
    feature_timeout : int, default: 3600
        Time budget for feature processing.
    use_wavelets : bool, default: False
        Whether to use wavelet transforms (useful for frequency data)

    # Interval analysis
    interval_analyse : bool, default: False
        Whether to use the IntervalAnalyser module.

    # Sequencing
    sequence : bool, default: False
        Whether to use the Sequencer module.
    seq_back : int or list of int, default: 1
        Input time indices.
        If int: includes that many samples backward.
        If list of int: includes all integers within the list.
    seq_forward : int or list of int, default: 1
        Output time indices.
        If int: includes that many samples forward.
        If list of int: includes all integers within the list.
    seq_shift : int, default: 0
        Shift input / output samples in time.
    seq_diff : {"none", "diff", "log_diff"}, default: "none"
        Difference the input and output.
    seq_flat : bool, default: True
        Whether to return a matrix (True) or a tensor (False).

    # Modelling
    standardize : bool, default: False
        Whether to standardize the input/output data.
    cv_shuffle : bool, default: True
        Whether to shuffle the samples during cross-validation.
    cv_splits : int, default: 10
        How many cross-validation splits to make.
    store_models : bool, default: False
        Whether to store all trained model files.

    # Grid search
    grid_search_timeout : int, default: 3600
        Time budget for grid search (in seconds).
    n_grid_searches : int, default: 3
        Run grid search for the best `n_grid_searches` (model, feature set) pairs from
        initial modelling.
    n_trials_per_grid_search : int, default: 250
        Maximal number of trials/candidates for each grid search.

    # Flags
    process_data : bool, default: True
        Whether to force data processing.
    no_dirs : bool, default: False
        Whether to create files.

    # Other
    kwargs: Any
        Swallows all arguments that are not accepted. Warnings are raised if not empty.
    """

    def __init__(
        self,
        # Main settings
        main_dir: str = "Auto_ML/",
        target: str = "target",
        name: str = "AutoML",
        version: int = 1,
        mode: str | None = None,
        objective: str | None = None,
        verbose: int = 1,
        logging_to_file: bool = False,
        logging_path: str | None = None,
        *,
        # Data processing
        missing_values: str = "zero",
        outlier_removal: str = "clip",
        z_score_threshold: int = 4,
        include_output: bool = False,
        # Balancing
        balance: bool = False,
        # Feature processing
        extract_features: bool = True,
        information_threshold: float = 0.999,
        feature_timeout: int = 3600,
        use_wavelets: bool = False,
        # Interval analysis
        interval_analyse: bool = False,
        # Sequencing
        sequence: bool = False,
        seq_back: int | list[int] = 1,
        seq_forward: int | list[int] = 1,
        seq_shift: int = 0,
        seq_diff: str = "none",
        seq_flat: bool = True,
        # Modelling
        standardize: bool = False,
        cv_shuffle: bool = True,
        cv_splits: int = 10,
        store_models: bool = False,
        # Grid search
        grid_search_timeout: int = 3600,
        n_grid_searches: int = 2,
        n_trials_per_grid_search: int = 250,
        # Flags
        process_data: bool = True,
        no_dirs: bool = False,
        # Other
        **kwargs,
    ):
        # Get init parameters for `self.settings`
        sig, init_locals = signature(self.__init__), locals()
        init_params = {
            param.name: init_locals[param.name] for param in sig.parameters.values()
        }
        del sig, init_locals

        # Initialize Logger
        super().__init__(verbose=verbose)
        if logging_path is None:
            logging_path = f"{Path(main_dir)}/AutoML.log"
        if logging_to_file:
            logging.add_file_handler(logging_path)

        # Input checks: validity
        if mode not in (None, "regression", "classification"):
            raise ValueError("Supported models: {'regression', 'classification', None}")
        if not 0 < information_threshold < 1:
            raise ValueError("Information threshold must be within (0, 1) interval.")

        # Input checks: advices
        if kwargs:
            warn(f"Got unexpected keyword arguments that are not handled: {kwargs}")
        if include_output and not sequence:
            warn("It is strongly advised NOT to include output without sequencing.")
        if interval_analyse and not standardize:
            warn(
                "Data needs to be normalized for the interval analyser, setting "
                "standardize = True."
            )
            standardize = True

        # Main settings
        self.main_dir = f"{Path(main_dir)}/"  # assert backslash afterwards
        self.target = target
        self.name = name
        self.version = version
        self.mode = mode
        self.objective = objective

        # Data processing
        self.missing_values = missing_values
        self.outlier_removal = outlier_removal
        self.z_score_threshold = z_score_threshold
        self.include_output = include_output

        # Balancing
        self.balance = balance

        # Feature processing
        self.extract_features = extract_features
        self.information_threshold = information_threshold
        self.feature_timeout = feature_timeout
        self.use_wavelets = use_wavelets

        # Interval analysis
        self.use_interval_analyser = interval_analyse

        # Sequencing
        self.sequence = sequence
        self.sequence_back = seq_back
        self.sequence_forward = seq_forward
        self.sequence_shift = seq_shift
        self.sequence_diff = seq_diff
        self.sequence_flat = seq_flat

        # Modelling
        self.standardize = standardize
        self.cv_shuffle = cv_shuffle
        self.cv_splits = cv_splits
        self.store_models = store_models

        # Grid search
        self.grid_search_timeout = grid_search_timeout
        self.n_grid_searches = n_grid_searches
        self.n_trials_per_grid_search = n_trials_per_grid_search

        # Flags
        self.process_data = process_data
        self.no_dirs = no_dirs

        # Set version
        self.version = version if version else 1

        # Store Pipeline Settings
        self.settings: dict[str, Any] = {"pipeline": init_params}

        # Objective & Scorer
        if self.objective is not None:
            if not isinstance(self.objective, str):
                raise ValueError("Objective needs to be a string.")
            self.scorer = metrics.get_scorer(self.objective)
        else:
            self.scorer = None

        # Required sub-classes
        self.data_processor = None
        self.data_sequencer = None
        self.interval_analyser = None
        self.feature_processor = None
        self.standardizer = None

        # Instance initiating
        self.best_model_: BasePredictor | None = None
        self.best_model_str_: str | None = None
        self.best_params_: dict[str, Any] | None = None
        self.best_feature_set_: str | None = None
        self.best_score_: float | None = None
        self.feature_sets_: dict[str, list[str]] | None = None
        self.results_: pd.DataFrame = pd.DataFrame(
            columns=[
                "feature_set",
                "score",
                "worst_case",
                "date",
                "model",
                "params",
                "time",
            ]
        )
        self.is_fitted_ = False

        # Monitoring
        self._prediction_time_: float | None = None
        self.main_predictors_: dict | None = None

    # User Pointing Functions
    def load(self):
        """
        Restores a pipeline from directory, given main_dir and version.
        """
        assert self.main_dir and self.version

        # Load settings
        settings_path = self.main_dir + "Settings.json"
        with open(settings_path, "r") as settings:
            self.load_settings(json.load(settings))

        # Load model
        model_path = self.main_dir + "Model.joblib"
        self.load_model(joblib.load(model_path))

    def load_settings(self, settings: dict):
        """
        Restores a pipeline from settings.

        Parameters
        ----------
        settings : dict
            Pipeline settings.
        """
        # Set parameters
        settings["pipeline"]["no_dirs"] = True
        settings["pipeline"]["main_dir"] = self.main_dir
        self.__init__(**settings["pipeline"])
        self.settings = settings
        self.best_model_str_ = settings.get("model")
        self.best_params_ = settings.get("params", {})
        self.best_score_ = settings.get("best_score")
        self.best_features_ = settings.get("features")
        self.best_feature_set_ = settings.get("feature_set")
        self.data_processor = DataProcessor().load_settings(settings["data_processing"])
        self.feature_processor = FeatureProcessor().load_settings(
            settings["feature_processing"]
        )
        self.standardizer = Standardizer().load_settings(
            settings.get("standardizing", {})
        )

    def load_model(self, model: BasePredictor):
        """
        Restores a trained model
        """
        assert type(model).__name__ == self.settings["model"]
        self.best_model_ = model
        self.is_fitted_ = True

    def fit(
        self,
        data_or_path: np.ndarray | pd.DataFrame | str | Path,
        target: np.ndarray | pd.Series | str | None = None,
        *,
        metadata: dict[str, dict] | None = None,
        model: str | list[str] | None = None,
        feature_set: str | list[str] | None = None,
    ):
        """
        Fit the full AutoML pipeline.
            1. Prepare data for training
            2. Train / optimize models
            3. Prepare Production Files
                Nicely organises all required scripts / files to make a prediction

        Parameters
        ----------
        data_or_path : np.ndarray or pd.DataFrame or str or Path
            Data or path to data. Propagated to `self.data_preparation`.
        target : np.ndarray or pd.Series or str
            Target data or column name. Propagated to `self.data_preparation`.
        *
        metadata : dict of {int : dict of {str : str or float}}, optional
            Metadata. Propagated to `self.data_preparation`.
        model : str or list of str, optional
            Constrain grid search and fitting conclusion to given model(s).
            Propagated to `self.model_training` and `self.conclude_fitting`.
        feature_set : str or list of str, optional
            Constrain grid search and fitting conclusion to given feature set(s).
            Propagated to `self.model_training` and `self.conclude_fitting`.
            Options: {rf_threshold, rf_increment, shap_threshold, shap_increment}
        params : dict, optional
            Constrain parameters for fitting conclusion.
            Propagated to `self.conclude_fitting`.
        """
        # Starting
        self.logger.info(f"\n\n*** Starting Amplo AutoML - {self.name} ***\n\n")

        # Reading data
        data = self._read_data(data_or_path, target, metadata=metadata)

        # Detect mode (classification / regression)
        self._mode_detector(data)
        self._set_subclasses()
        assert (
            self.mode
            and self.objective
            and self.data_processor
            and self.data_sequencer
            and self.interval_analyser
            and self.feature_processor
            and self.standardizer
        )

        # Preprocess Data
        data = self.data_processor.fit_transform(data)

        # Sequence
        if self.sequence:
            data = self.data_sequencer.fit_transform(data)

        # Extract and select features
        data = self.feature_processor.fit_transform(data)
        self.feature_sets_ = self.feature_processor.feature_sets_

        # Standardize
        if self.standardize:
            data = self.standardizer.fit_transform(data)

        # Interval-analyze data
        if (
            self.use_interval_analyser
            and len(data.index.names) == 2
            and self.mode == "classification"
        ):
            data = self.interval_analyser.fit_transform(data)

        # Model Training #
        ##################
        # TODO: add model limitation
        for feature_set, cols in self.feature_sets_.items():
            self.logger.info(f"Fitting modeller on: {feature_set}")
            feature_data: pd.DataFrame = data[cols + [self.target]]
            results_ = Modeller(
                target=self.target,
                mode=self.mode,
                cv=self.cv,
                objective=self.objective,
                verbose=self.verbose,
            ).fit(feature_data)
            results_["feature_set"] = feature_set
            self.results_ = self.sort_results(pd.concat([results_, self.results_]))

        # Optimize Hyper parameters
        for model, feature_set in self.grab_grid_search_iterations():
            # TODO: implement models limitations
            assert feature_set in self.feature_sets_
            self.logger.info(
                f"Starting Hyper Parameter Optimization for {model} on "
                f"{feature_set} features ({len(data)} samples, "
                f"{len(self.feature_sets_[feature_set])} features)"
            )
            results_ = OptunaGridSearch(
                get_model(model),
                target=self.target,
                timeout=self.grid_search_timeout,
                cv=self.cv,
                n_trials=self.n_trials_per_grid_search,
                scoring=self.objective,
                verbose=self.verbose,
            ).fit(data)
            results_["feature_set"] = feature_set
            self.results_ = self.sort_results(
                pd.concat([self.results_, results_], ignore_index=True)
            )

        # Storing model
        self.store_best(data)

        # Observe
        self.settings["data_observer"] = DataObserver().observe(
            data, self.mode, self.target, self.data_processor.dummies_
        )
        self.settings["model_observer"] = ModelObserver().observe(
            self.best_model_, data, self.target, self.mode  # type: ignore
        )

        # Finish
        self.is_fitted_ = True
        self.logger.info("All done :)")
        logging.del_file_handlers()

    def grab_grid_search_iterations(self) -> list[tuple[str, str]]:
        iterations = []
        for i in range(self.n_grid_searches):
            row = self.results_.iloc[i]
            iterations.append((row["model"], row["feature_set"]))
        return iterations

    def store_best(self, data: pd.DataFrame):
        # TODO implement models limitations
        assert (
            self.feature_sets_
            and self.scorer
            and self.mode
            and self.data_processor
            and self.data_sequencer
            and self.feature_processor
            and self.standardizer
        )

        # Gather best results_
        self.best_score_ = self.results_.iloc[0]["worst_case"]
        self.best_model_str_ = self.results_.iloc[0]["model"]
        self.best_feature_set_ = self.results_.iloc[0]["feature_set"]
        self.best_features_ = self.feature_sets_.get(self.best_feature_set_, [])
        parsed_params = io.parse_json(self.results_.iloc[0]["params"])
        assert isinstance(parsed_params, dict)
        self.best_params_ = parsed_params

        # Train model on all training data
        self.best_model_ = get_model(self.best_model_str_)
        self.best_model_.set_params(**self.best_params_)
        self.best_model_.fit(data[self.best_features_], data[self.target])

        # Prune Data Processor
        required_features = get_required_columns(
            self.feature_processor.feature_sets_[self.best_feature_set_],
            self.feature_processor.numeric_cols_,
        )
        self.data_processor.prune_features(required_features)

        # Update pipeline settings
        self.settings["version"] = self.version
        self.settings["pipeline"]["verbose"] = self.verbose
        self.settings["model"] = self.best_model_str_
        self.settings["params"] = self.best_params_
        self.settings["feature_set"] = self.best_feature_set_
        self.settings["features"] = self.best_features_
        self.settings["data_processing"] = self.data_processor.get_settings()
        self.settings["feature_processing"] = self.feature_processor.get_settings()
        if self.standardize:
            self.settings["standardizing"] = self.standardizer.get_settings()
        self.settings["best_score"] = self.best_score_
        self.settings["amplo_version"] = (
            amplo.__version__ if hasattr(amplo, "__version__") else "dev"  # type: ignore
        )

        # Validation
        validator = ModelValidator(
            target=self.target,
            cv=self.cv,
            verbose=self.verbose,
        )
        self.settings["validation"] = validator.validate(
            model=self.best_model_, data=data, mode=self.mode
        )

        # Return if no_dirs flag is set
        if self.no_dirs:
            return

        # Create directory
        if not os.path.exists(self.main_dir):
            os.makedirs(self.main_dir)

        # Save model & settings
        joblib.dump(self.best_model_, self.main_dir + "Model.joblib")
        with open(self.main_dir + "Settings.json", "w") as settings:
            json.dump(self.settings, settings, indent=4, cls=io.NpEncoder)

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        if not self.is_fitted_:
            raise ValueError("Pipeline not yet fitted.")
        assert self.data_processor and self.feature_processor and self.standardizer

        # Process data
        data = self.data_processor.transform(data)

        # Sequence
        if self.sequence:
            warn("Sequencer is temporarily disabled.", DeprecationWarning)

        # Convert Features
        data = self.feature_processor.transform(
            data, feature_set=self.settings["feature_set"]
        )

        # Standardize
        if self.standardize:
            data = self.standardizer.transform(data)

        # Output
        if not self.include_output and self.target in data:
            data = data.drop(self.target, axis=1)

        # Return
        return data

    def predict(self, data: pd.DataFrame) -> pd.Series:
        """
        Full script to make predictions. Uses 'Production' folder with defined or
        latest version.

        Parameters
        ----------
        data : pd.DataFrame
            Data to do prediction on.
        """
        start_time = time.time()
        if not self.is_fitted_:
            raise ValueError("Pipeline not yet fitted.")
        self.logger.info(
            f"Predicting with {type(self.best_model_).__name__}, v{self.version}"
        )

        # Convert
        data = self.transform(data)

        # Predict
        assert self.best_model_
        predictions = self.best_model_.predict(data)

        # Convert
        if self.mode == "regression" and self.standardize:
            assert self.standardizer
            predictions = self.standardizer.reverse(predictions, column=self.target)
        elif self.mode == "classification":
            assert self.data_processor
            predictions = self.data_processor.decode_labels(predictions)

        # Stop timer
        self._prediction_time_ = (time.time() - start_time) / len(data) * 1000

        # Calculate main predictors
        self._get_main_predictors(data)

        return predictions

    def predict_proba(self, data: pd.DataFrame) -> np.ndarray:
        """
        Returns probabilistic prediction, only for classification.

        Parameters
        ----------
        data : pd.DataFrame
            Data to do prediction on.
        """
        start_time = time.time()
        if not self.is_fitted_:
            raise ValueError("Pipeline not yet fitted.")
        if self.mode != "classification":
            raise ValueError("Predict_proba only available for classification")
        if not hasattr(self.best_model_, "predict_proba"):
            raise ValueError(
                f"{type(self.best_model_).__name__} has no attribute predict_proba"
            )
        self.logger.info(
            f"Predicting with {type(self.best_model_).__name__}, v{self.version}"
        )

        # Convert data
        data = self.transform(data)

        # Predict
        prediction = self.best_model_.predict_proba(data)  # type: ignore -- asserted

        # Stop timer
        self._prediction_time_ = (time.time() - start_time) / len(data) * 1000

        # Calculate main predictors
        self._get_main_predictors(data)

        return prediction

    # Fit functions
    def _read_data(
        self,
        data_or_path: np.ndarray | pd.DataFrame | str | Path,
        target: list | tuple | np.ndarray | pd.Series | str | Path | None = None,
        *,
        metadata: dict[str, dict] | None = None,
    ) -> pd.DataFrame:
        """
        Read and validate data.

        Notes
        -----
        The required parameters depend on the input parameter types.

        When ``target`` is None, it is set to ``self.target`` or "target" otherwise.

        When ``data_or_path`` is path-like, then the parameters ``target`` and
        ``metadata`` are not required.
        Otherwise, when ``data_or_path`` is array-like, it either must contain a column
        name as the ``target`` parameter indicates or ``target`` must also be an
        array-like object with the same length as ``data_or_path``.

        Note: There's three combinations of data_or_path and target
        1. if data_or_path = pd.DataFrame, target = pd.Series | None | str
        2. if data_or_path = np.ndarray, target = np.ndarray | pd.Series
        3. if data_or_path = path | str, target = path | str | None

        Parameters
        ----------
        data_or_path : np.ndarray or pd.DataFrame or str or Path
            Data or path to data.
        target : np.ndarray or pd.Series or str
            Target data or column name or directory name
        *
        metadata : dict of {int : dict of {str : str or float}}, optional
            Metadata.

        Returns
        -------
        Pipeline
            The same object but with injected data.
        """
        # 1. if data_or_path = pd.DataFrame, target = ArrayLike | str | None
        if isinstance(data_or_path, pd.DataFrame):
            data = data_or_path
            # If it's a series, we check index and take the name
            if isinstance(target, pd.Series):
                if not all(data.index == target.index):
                    warn(
                        "Indices of data and target don't match. Target index will be "
                        "overwritten by data index."
                    )
                    target.index = data.index
                if target.name and self.target != target.name:
                    warn(
                        "Provided target series has a different name than initialized "
                        "target. Using series name."
                    )
                    self.target = str(target.name)
            # Then for arraylike, we check length and make sure target is not in data
            if isinstance(target, (list, tuple, pd.Series, np.ndarray)):
                if len(data) != len(target):
                    raise ValueError("Length of target and data don't match.")
                if self.target in data and (data[self.target] != target).any():
                    raise ValueError(
                        f"The column '{self.target}' column already exists in `data` "
                        f"but has different values."
                    )
                data[self.target] = target
            # If it's a string, we check its presence and update self.target
            elif isinstance(target, str):
                if target not in data:
                    raise ValueError("Provided target column not present in data.")
                self.target = target
            # If it's none, self.target is taken from __init__
            elif isinstance(target, type(None)):
                if self.target not in data:
                    raise ValueError("Initialized target column not present in data.")
            else:
                raise NotImplementedError(
                    "When data_or_path is a DataFrame, target needs to "
                    "be a Series, str or None"
                )

        # 2. if data_or_path = np.ndarray, target = ArrayLike
        elif isinstance(data_or_path, np.ndarray):
            if not isinstance(target, (np.ndarray, pd.Series, list, tuple)):
                raise NotImplementedError(
                    "If data is ndarray, target should be ArrayLike."
                )
            if len(data_or_path) != len(target):
                raise ValueError("Length of target and data don't match.")
            if isinstance(target, pd.Series):
                data = pd.DataFrame(data_or_path, index=target.index)
                if target.name:
                    self.target = str(target.name)
            else:
                data = pd.DataFrame(data_or_path)
            data[self.target] = target

        # 3. if data_or_path = path | str, target = path | str | None
        elif isinstance(data_or_path, (str, Path)):
            if isinstance(target, (str, Path)):
                self.target = str(target)
            elif not isinstance(target, type(None)):
                raise ValueError(
                    "Target must be string | Path | None when `data_or_path` is a "
                    "path-like object."
                )
            if metadata:
                warn(
                    "Parameter `metadata` is ignored when `data_or_path` is a "
                    "path-like object."
                )
            data, metadata = io.merge_logs(parent=data_or_path, target=self.target)

        # 4. Error.
        else:
            raise NotImplementedError(
                "Supported data_or_path types: pd.DataFrame | np.ndarray | Path | str"
            )
        assert isinstance(data, pd.DataFrame)

        # Clean target name
        clean_target = clean_feature_name(self.target)
        data = data.rename(columns={self.target: clean_target})
        self.target = clean_target

        # Finish
        self.settings["file_metadata"] = metadata or {}

        return data

    def has_new_training_data(self):
        # TODO: fix a better solution for this
        return True

    def _mode_detector(self, data: pd.DataFrame):
        """
        Detects the mode (Regression / Classification)

        parameters
        ----------
        data : pd.DataFrame
        """
        # Only run if mode is not provided
        if self.mode in ("classification", "regression"):
            return

        # Classification if string
        labels = data[self.target]
        if labels.dtype == str or labels.nunique() < 0.1 * len(data):
            self.mode = "classification"
            self.objective = self.objective or "neg_log_loss"

        # Else regression
        else:
            self.mode = "regression"
            self.objective = self.objective or "neg_mean_absolute_error"

        # Set scorer
        self.scorer = metrics.get_scorer(self.objective)

        # Copy to settings
        self.settings["pipeline"]["mode"] = self.mode
        self.settings["pipeline"]["objective"] = self.objective

        # Logging
        self.logger.info(
            f"Setting mode to {self.mode} & objective to {self.objective}."
        )

    def _set_subclasses(self):
        """
        Simple function which sets subclasses. This cannot be done
        during class initialization due to certain attributes which
        are data dependent. Data is only known at calling .fit().
        """
        assert self.mode
        self.data_processor = DataProcessor(
            target=self.target,
            drop_datetime=True,
            include_output=True,
            missing_values=self.missing_values,
            outlier_removal=self.outlier_removal,
            z_score_threshold=self.z_score_threshold,
        )
        self.data_sequencer = Sequencer(
            target=self.target,
            back=self.sequence_back,
            forward=self.sequence_forward,
            shift=self.sequence_shift,
            diff=self.sequence_diff,
        )
        self.interval_analyser = IntervalAnalyser(target=self.target)
        self.feature_processor = FeatureProcessor(
            target=self.target,
            mode=self.mode,
            is_temporal=None,
            use_wavelets=self.use_wavelets,
            extract_features=self.extract_features,
            collinear_threshold=self.information_threshold,
            verbose=self.verbose,
        )
        self.standardizer = Standardizer(target=self.target, mode=self.mode)

    # Getter Functions / Properties
    @property
    def cv(self):
        """
        Gives the Cross Validation scheme

        Returns
        -------
        cv : KFold or StratifiedKFold
            The cross validator
        """
        # Regression
        if self.mode == "regression":
            return KFold(
                n_splits=self.cv_splits,
                shuffle=self.cv_shuffle,
                random_state=83847939 if self.cv_shuffle else None,
            )

        # Classification
        if self.mode == "classification":
            return StratifiedKFold(
                n_splits=self.cv_splits,
                shuffle=self.cv_shuffle,
                random_state=83847939 if self.cv_shuffle else None,
            )
        else:
            raise NotImplementedError("Unknown Mode.")

    # Support Functions
    @staticmethod
    def sort_results(results_: pd.DataFrame) -> pd.DataFrame:
        return results_.sort_values("worst_case", ascending=False)

    def _get_main_predictors(self, data):
        """
        Using Shapely Additive Explanations, this function calculates the main
        predictors for a given prediction and sets them into the class' memory.
        """
        # shap.TreeExplainer is not implemented for all models. So we try and fall back
        # to the feature importance given by the feature processor.
        # Note that the error would be raised when calling `TreeExplainer(best_model_)`.
        try:
            # Get shap values
            best_model_ = self.best_model_
            if type(best_model_).__module__.startswith("amplo"):
                best_model_ = best_model_.model  # type: ignore
            # Note: The error would be raised at this point.
            #  So we have not much overhead.
            shap_values = np.array(TreeExplainer(best_model_).shap_values(data))

            # Average over classes if necessary
            if shap_values.ndim == 3:
                shap_values = np.mean(np.abs(shap_values), axis=0)

            # Average over samples
            shap_values = np.mean(np.abs(shap_values), axis=0)
            shap_values /= shap_values.sum()  # normalize to sum up to 1
            idx_sort = np.flip(np.argsort(shap_values))

            # Set class attribute
            main_predictors = {
                col: score
                for col, score in zip(data.columns[idx_sort], shap_values[idx_sort])
            }

        except Exception:  # the exception can't be more specific  # noqa
            # Get shap feature importance
            assert self.feature_processor
            fi = self.feature_processor.feature_importance_.get("rf", {})

            # Use only those columns that are present in the data
            main_predictors = {}
            missing_columns = []
            for col in data:
                if col in fi:
                    main_predictors[col] = fi[col]
                else:
                    missing_columns.append(col)

            if missing_columns:
                warn(
                    f"Some data column names are missing in the shap feature "
                    f"importance dictionary: {missing_columns}"
                )

        # Some feature names are obscure since they come from the feature processing
        # module. Here, we relate the feature importance back to the original features.
        translation = translate_features(list(main_predictors))
        scores = {}
        for key, features in translation.items():
            for feat in features:
                scores[feat] = scores.get(feat, 0.0) + main_predictors[key]
        # Normalize
        total_score = np.sum(list(scores.values()))
        for key in scores:
            scores[key] /= total_score

        # Set attribute
        self.main_predictors_ = scores

        # Add to settings: [{"feature": "feature_name", "score": 1}, ...]
        scores_df = pd.DataFrame({"feature": scores.keys(), "score": scores.values()})
        scores_df.sort_values("score", ascending=False, inplace=True)
        self.settings["main_predictors"] = scores_df.to_dict("records")

        return scores



#  Copyright (c) 2022 by Amplo.

"""
Feature processor for extracting no features at all.
"""


import pandas as pd

from amplo.automl.feature_processing._base import BaseFeatureExtractor

__all__ = ["NopFeatureExtractor"]


class NopFeatureExtractor(BaseFeatureExtractor):
    """
    Feature processor for extracting no features.

    Each input column will be accepted as a feature.
    """

    def fit(self, data: pd.DataFrame):
        self.reset()
        x, _ = self._check_data(data, require_y=False)

        # Fitting: accept each feature/column
        self.add_features(x)
        self._is_fitted = True

        return self

    def fit_transform(self, data: pd.DataFrame) -> pd.DataFrame:
        return self.fit(data).transform(data)

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        self.check_is_fitted()
        x, _ = self._check_data(data, require_y=False)

        # Transforming: select 'fitted' features
        return x[self.features_]

    def transform_target(self, y: pd.Series) -> pd.Series:
        self.check_is_fitted()
        return self._check_y(y, copy=False)



#  Copyright (c) 2022 by Amplo.

import ast
import re

import numpy as np
import pandas as pd
import pytest

from amplo.automl import DataProcessor
from amplo.observation._base import ProductionWarning
from amplo.observation.data import DataObserver
from tests import find_first_warning_of_type


class TestDataObserver:
    @pytest.mark.parametrize("index_type", ["single", "multi"])
    def test_monotonic_columns(self, index_type):
        size = 100
        monotonic_incr = 4.2 * np.arange(-10, size - 10)
        monotonic_decr = 6.3 * np.arange(-3, size - 3)[::-1]
        constants = np.zeros(size)
        random = np.random.normal(size=size)

        # Prepare data
        y1 = random.reshape(-1)  # does not matter
        if index_type == "single":
            data = pd.DataFrame(
                {
                    "inc": monotonic_incr,
                    "dec": monotonic_decr,
                    "con": constants,
                    "random": random,
                    "y1": y1,
                }
            )
        elif index_type == "multi":
            data = pd.DataFrame(
                {
                    "inc": np.concatenate([monotonic_incr, monotonic_incr], axis=0),
                    "dec": np.concatenate([monotonic_decr, monotonic_decr], axis=0),
                    "con": np.concatenate([constants, constants], axis=0),
                    "random": np.concatenate([random, random], axis=0),
                    "y1": np.concatenate([y1, y1], axis=0),
                },
                index=pd.MultiIndex.from_product([[0, 1], range(size)]),
            )
        else:
            raise ValueError("Invalid index_type.")

        # Add nan to first and random
        data.iloc[1, 0] = np.nan

        # Observe)
        with pytest.warns(ProductionWarning) as record:
            DataObserver().check_monotonic_columns(data)
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        monotonic_cols = ast.literal_eval(re.search(r"\[.*]", msg).group(0))  # type: ignore
        assert set(monotonic_cols) == {
            "dec",
            "inc",
        }, "Wrong monotonic columns identified."

    def test_minority_sensitivity(self):
        # Setup
        data = pd.DataFrame(
            {
                "random": np.random.normal(size=100),
                "x": np.concatenate((np.zeros(2), np.random.normal(100, 1, 98))),
                "y": np.concatenate((np.zeros(5), np.ones(95))),
            }
        )
        data.iloc[1, 0] = np.nan

        # Observe
        with pytest.warns(ProductionWarning) as record:
            DataObserver().check_minority_sensitivity(data, target="y")
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        sensitive_cols = ast.literal_eval(re.search(r"\[.*]", msg).group(0))  # type: ignore
        assert sensitive_cols == ["x"], "Wrong minority sensitive columns identified."

    def test_categorical_mismatch(self):
        """
        Detects whether there is an issue with categorical variables being too similar.

        Note: it does not detect "New York" vs. "new-york", as this is taken care off by
        cleaning feature names.
        """
        # Setup
        data = pd.DataFrame(
            {
                "cat_a": np.array(["New York"] * 50 + ["new-york"] * 50),
                "cat_b": np.array(["Something"] * 50 + ["Somethign"] * 50),
                "normal": np.random.normal(100, 5, 100),
                "y": np.concatenate((np.zeros(5), np.ones(95))),
            }
        )

        # Add nan
        data.iloc[0, 0] = np.nan

        # Observe
        dp = DataProcessor()
        dp.fit_transform(data)
        with pytest.warns(ProductionWarning) as record:
            DataObserver().check_categorical_mismatch(dp.dummies_)
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        sensitive_cols = ast.literal_eval(re.search(r"\[.*]", msg).group(0))  # type: ignore
        assert (
            len(sensitive_cols) == 1 and "cat_b" in sensitive_cols[0]
        ), "Wrong categorical mismatch columns identified."

    def test_extreme_values(self):
        # Setup
        data = pd.DataFrame(
            {
                "normal": np.random.normal(size=100),
                "linear": np.linspace(1000, 10000, 100),
                "y": np.concatenate((np.zeros(5), np.ones(95))),
            }
        )

        # Add nan
        data.iloc[0, 0] = np.nan

        # Observe
        with pytest.warns(ProductionWarning) as record:
            DataObserver().check_extreme_values(data)
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        extreme_cols = ast.literal_eval(re.search(r"\[.*]", msg).group(0))  # type: ignore
        assert extreme_cols == [
            "linear"
        ], "Wrong minority sensitive columns identified."

    def test_label_issues(self):
        # Setup
        data = pd.DataFrame(
            {
                "x": np.hstack(
                    (
                        np.random.normal(-2, 0.1, size=50),
                        np.random.normal(2, 0.1, size=50),
                    )
                ),
                "y": np.hstack((np.ones(50), np.zeros(49), 1)),
            }
        )

        # Observe
        with pytest.warns(ProductionWarning) as record:
            DataObserver().check_label_issues(data, mode="classification", target="y")
        msg = str(find_first_warning_of_type(ProductionWarning, record).message)
        label_issues = ast.literal_eval(re.search(r"\[.*]", msg).group(0))  # type: ignore
        assert label_issues == [99], "Wrong label issues identified."



#  Copyright (c) 2022 by Amplo.
from __future__ import annotations

from warnings import warn

import numpy as np
import pandas as pd

from amplo.base import LoggingMixin


class Sequencer(LoggingMixin):
    # todo implement fractional differencing

    def __init__(
        self,
        target: str,
        back: int | list[int] = 1,
        forward: int | list[int] = 1,
        shift=0,
        diff="none",
        verbose: int = 1,
    ):
        """Sequences and differentiates data.

        Parameters
        ----------
        target : str
            Target column
        back : list of int or int
            Input indices (see also Notes).
        forward : list of int or int
            Output indices (see also Notes).
        shift : int
            Indicates the shift between input and output.
        diff : str
            Depicts the differencing algorithm. Options: `none`, `diff` and `log_diff`.

        Notes
        -----
        For the ``back`` and ``forward`` parameter, behavior changes depending on its
        type:
            - If it's an **integer** dtype, indices include that many samples backward
            or forward, respectively.
            - If it's an **iterable** dtype, indices include all integers within the
            iterable.

        The indices of ``back`` and ``forward`` start from ``0``. Therefore, if the
        output is included in the input, having ``forward=4`` will result in predicting
        the output for ``[t, t+1, t+2, t+3]`` and having ``forward=[4]`` will result in
        making a ``t+4`` prediction. The same rules apply to the ``forward`` parameter.
        """
        super().__init__(verbose=verbose)
        # Tests
        if diff not in [
            "none",
            "diff",
            "log_diff",
        ]:
            raise ValueError("Diff needs to be none, diff or log_diff.")
        if isinstance(back, int):
            if back <= 0:
                raise ValueError("'back' arg needs to be strictly positive.")
        else:
            if not all([x >= 0 for x in back]):
                raise ValueError("All integers in 'back' need to be positive.")
            if not all([x > 0 for x in np.diff(back)]):
                raise ValueError(
                    "All integers in 'back' need to be monotonically increasing."
                )
        if isinstance(forward, int) and forward < 0:
            raise ValueError("'forward' arg needs to be positive")
        elif isinstance(forward, list):
            if not all([x >= 0 for x in forward]):
                raise ValueError("All integers in 'forward' need to be positive.")
            if not all([x > 0 for x in np.diff(forward)]):
                raise ValueError(
                    "All integers in 'forward' need to be monotonically increasing."
                )
        if diff != "none" and isinstance(back, int) and back <= 1:
            raise ValueError("With differencing, back needs to be at least 2.")
        if diff != "none" and isinstance(forward, int) and forward <= 1:
            raise ValueError("With differencing, forward needs to be at least 2.")

        # Note static args
        self.target = target
        self.shift = shift
        self.diff = diff
        self.samples_ = 0
        self.input_constant_ = 0
        self.output_constant_ = 0

        # Parse args
        if isinstance(back, int):
            back_arr = np.linspace(0, back - 1, back).astype("int")
        elif isinstance(back, list):
            back_arr = np.array(back)
        if isinstance(forward, int):
            forward_arr = np.linspace(0, forward - 1, forward).astype("int")
        elif isinstance(forward, list):
            forward_arr = np.array(forward)

        # Index Vectors
        assert len(back_arr) > 0
        assert len(forward_arr) > 0
        self.input_indices_ = back_arr
        self.output_indices_ = forward_arr

        # Differencing vectors
        if diff != "none":
            # In case first is 0, we won't difference with -1, therefore, we add & skip
            # the first
            if self.input_indices_[0] == 0:
                self.input_diff_indices_ = self.input_indices_[:-1]
                self.input_indices_ = self.input_indices_[1:]
            else:
                # However, if first is nonzero, we can keep all and roll them, changing
                # first one to 0
                self.input_diff_indices_ = np.roll(self.input_indices_, 1)
                self.input_diff_indices_[0] = 0

            # Same for output
            if self.output_indices_[0] == 0:
                self.output_diff_indices_ = self.output_indices_[:-1]
                self.output_indices_ = self.output_indices_[1:]
            else:
                self.output_diff_indices_ = np.roll(self.output_indices_, 1)
                self.output_diff_indices_[0] = 0

        # Number of sequence steps
        self.n_input_steps_ = len(self.input_indices_)
        self.n_output_steps_ = len(self.output_indices_)

        # Maximum steps
        self.max_input_step_ = max(self.input_indices_)
        self.max_output_step_ = max(self.output_indices_)

    def fit_transform(self, data: pd.DataFrame, flat=True) -> pd.DataFrame:
        """
        Sequences input / outputs dataframe / numpy array.

        parameters
        ----------
        data : pd.DataFrame
        flat : bool, default=True
            Whether to return a 2d matrix or 3d
        """
        # Split data
        if self.target not in data:
            raise ValueError(f"Target column {self.target} missing.")
        y = data[self.target]
        x = data.drop(self.target, axis=1)

        # Initials
        input_keys = x.keys()
        output_keys = y.keys()

        # If flat return
        if flat:
            # No Differencing
            if self.diff == "none":
                # Input
                for lag in self.input_indices_:
                    keys = [key + "_" + str(lag) for key in input_keys]
                    x[keys] = x[input_keys].shift(lag)

                # Output
                for shift in self.output_indices_:
                    keys = [key + "_" + str(shift) for key in output_keys]
                    y[keys] = y[output_keys].shift(-shift)

            # With differencing
            elif self.diff[-4:] == "diff":
                # Input
                for lag in self.input_indices_:
                    # Shifted
                    keys = [key + "_" + str(lag) for key in input_keys]
                    x[keys] = x[input_keys].shift(lag)

                    # differentiated
                    d_keys = [key + "_d_" + str(lag) for key in input_keys]
                    x[d_keys] = x[input_keys].shift(lag) - x[input_keys]

                # Output
                for shift in self.output_indices_:
                    # Only differentiated
                    keys = [key + "_" + str(shift) for key in output_keys]
                    y[keys] = y[output_keys].shift(shift) - y[output_keys]
            else:
                raise NotImplementedError("Unknown differencing algorithm.")

            # Drop _0 (same as original)
            x = x.drop([key for key in x.keys() if "_0" in key], axis=1)
            y = y.drop([key for key in y.keys() if "_0" in key], axis=1)

            # Return (first lags are NaN, last shifts are NaN
            x = x.iloc[lag : -shift if shift > 0 else None]  # type: ignore
            y = y.iloc[lag : -shift if shift > 0 else None]  # type: ignore
            x[self.target] = y
            return x
        else:
            raise NotImplemented(
                "Technically, we could use _convert_numpy, ",
                "but returning 3d is not supported by the pipeline.",
            )

    def _convert_numpy(
        self, x: np.ndarray, y: np.ndarray, flat=True
    ) -> tuple[np.ndarray, np.ndarray]:
        # Initializations
        if x.ndim == 1:
            x = x.reshape((-1, 1))
        if y.ndim == 1:
            y = y.reshape((-1, 1))
        # Samples, interval minus sequence length (maxIn+maxOutPlus+shift)
        self.samples_ += (
            len(x) - self.max_input_step_ - self.max_output_step_ - self.shift
        )
        features = len(x[0])
        input_sequence = np.zeros((self.samples_, self.n_input_steps_, features))
        output_sequence = np.zeros((self.samples_, self.n_output_steps_))

        # Sequence
        if self.diff == "none":
            for i in range(self.samples_):
                input_sequence[i] = x[i + self.input_indices_]
                output_sequence[i] = y[
                    i + self.max_input_step_ + self.shift + self.output_indices_
                ].reshape((-1))

        elif self.diff[-4:] == "diff":

            # Take log for log_diff
            if self.diff == "log_diff":
                if np.min(x) < 1e-3:
                    self.input_constant_ = abs(np.min(x) * 1.001) + 1e-3
                    warn(
                        f"Small or negative input values found, adding a constant "
                        f"{self.input_constant_:.2e} to input"
                    )
                if np.min(y) < 1e-3:
                    self.output_constant_ = abs(np.min(y) * 1.001) + 1e-3
                    warn(
                        f"Small or negative output values found, adding a constant "
                        f"{self.output_constant_:.2e} to output"
                    )
                x = np.log(x + self.input_constant_)
                y = np.log(y + self.output_constant_)

            # Finally create the difference vector
            for i in range(self.samples_):
                input_sequence[i] = (
                    x[i + self.input_indices_] - x[i + self.input_diff_indices_]
                )
                output_sequence[i] = (
                    y[i + self.max_input_step_ + self.shift + self.output_indices_]
                    - y[
                        i
                        + self.max_input_step_
                        + self.shift
                        + self.output_diff_indices_
                    ]
                ).reshape((-1))

        # Return
        if flat:
            return (
                input_sequence.reshape((self.samples_, self.n_input_steps_ * features)),
                output_sequence,
            )
        else:
            return input_sequence, output_sequence

    def revert(self, seq_y, y_start):
        """
        Reverts the sequenced vector back.

        Useful for sequenced predictions in production.

        :param seq_y: The sequenced / predicted signal,
        :param y_start: Starting vector, always necessary when differentiated.
        :return: y: normal predicted signal, de-sequenced.
        """
        if not len(seq_y.shape) == 2 and seq_y.shape[1] == self.n_output_steps_:
            raise ValueError("revert() only suitable for output.")

        # Initiate
        y = np.zeros((self.n_output_steps_, len(seq_y) + len(y_start)))

        # de-sequence if diff
        if self.diff == "diff":
            for i in range(self.n_output_steps_):
                y[i, : len(y_start)] = y_start
                for j in range(len(seq_y)):
                    y[i, j + self.output_indices_[i]] = (
                        y[0, j + self.output_diff_indices_[i]] + seq_y[j, i]
                    )
            return y

        # de-sequence if log_diff (take log of start, add sequence log, then take exp
        # (as x=exp(log(x)))
        elif self.diff == "log_diff":
            for i in range(self.n_output_steps_):
                y[i, : len(y_start)] = np.log(y_start + self.output_constant_)
                for j in range(len(seq_y)):
                    y[i, j + self.output_indices_] = (
                        y[0, j + self.output_diff_indices_] + seq_y[j, i]
                    )
                return np.exp(y) - self.output_constant_

        else:
            if self.diff == "none":
                raise ValueError(
                    "With differencing set to none, reverting is not necessary."
                )
            else:
                raise ValueError("Differencing method not implemented.")



#  Copyright (c) 2022 by Amplo.

from pathlib import Path

import pytest

from tests import rmtree

__all__ = ["TestAPI"]


class TestAPI:
    sync_dir = Path("./test_dir")

    @pytest.fixture(autouse=True)
    def rmtree_sync_dir(self):
        rmtree(self.sync_dir)
        yield
        rmtree(self.sync_dir)



#  Copyright (c) 2022 by Amplo.

from copy import deepcopy

from amplo.classification._base import BaseClassifier


class PartialBoostingClassifier(BaseClassifier):
    """
    Amplo wrapper for classification boosting models.

    The number of estimators being used in the prediction are limited.

    Parameters
    ----------
    model
        Boosting model to wrap.
    step : int
        Number of iterations/estimators to limit the model on predictions.
    verbose : {0, 1, 2}
        Verbose logging.
    """

    _SUPPORTED_MODELS = [
        "AdaBoostClassifier",
        "GradientBoostingClassifier",
        "LGBMClassifier",
        "XGBClassifier",
        "CatBoostClassifier",
    ]

    def __init__(self, model, step, verbose=0):
        model = deepcopy(model)

        model_name = type(model).__name__
        if model_name not in self._SUPPORTED_MODELS:
            raise ValueError(f"Unsupported model {model_name}")
        if model_name in ("AdaBoostClassifier", "GradientBoostingClassifier"):
            model.estimators_ = model.estimators_[:step]

        super().__init__(model=model, verbose=verbose)
        self.step = int(step)

    def _get_prediction_kwargs(self):
        model_name = type(self.model).__name__
        if model_name in ("AdaBoostClassifier", "GradientBoostingClassifier"):
            return {}
        elif model_name == "LGBMClassifier":
            return {"num_iterations": self.step}
        elif model_name == "XGBClassifier":
            return {"iteration_range": (0, self.step)}
        elif model_name == "CatBoostClassifier":
            return {"ntree_end": self.step}
        else:
            raise AttributeError(f"Unsupported model {model_name}")

    def _predict(self, x, y=None, **kwargs):
        return self.model.predict(x, **kwargs, **self._get_prediction_kwargs())

    def predict_proba(self, x, **kwargs):
        self.check_is_fitted()
        return self.model.predict_proba(x, **kwargs, **self._get_prediction_kwargs())

    @classmethod
    def n_estimators(cls, model):
        model_name = type(model).__name__
        if model_name in ("AdaBoostClassifier", "GradientBoostingClassifier"):
            return len(model.estimators_)
        elif model_name in ("LGBMClassifier", "XGBClassifier"):
            return model.model.n_estimators
        elif model_name == "CatBoostClassifier":
            return model.model.tree_count_
        else:
            raise ValueError(f"Unsupported model {model_name}")



#  Copyright (c) 2022 by Amplo.

from copy import deepcopy

import numpy as np
import pandas as pd
import pytest

from amplo.automl.feature_processing._base import (
    BaseFeatureExtractor,
    BaseFeatureProcessor,
    sanitize_dataframe,
    sanitize_series,
)


@pytest.fixture(scope="function")
def make_good_and_corrupt_classification(make_x_y):
    # Make corrupted data
    x_corrupt, y_corrupt = make_x_y
    x_corrupt.iloc[0, 0] = np.nan
    y_corrupt.iloc[0] = np.nan
    x_corrupt.columns = list(range(len(x_corrupt.columns)))
    y_corrupt.name = None

    # Make sanitized data
    x_good = sanitize_dataframe(x_corrupt)
    x_good.columns = list(map(str, x_good.columns))
    y_good = sanitize_series(y_corrupt)
    y_good.name = "target"

    yield x_good, x_corrupt, y_good, y_corrupt


class TestFunctions:
    def test_sanitize_series_and_dataframe(self):
        frame = pd.DataFrame(
            {
                "non_numeric": list("abcd"),
                "floats": [123, 1e12 + 1, -1e12 - 1, np.nan],
                "ints": [123, 1e12 + 1, -1e12 - 1, 0],
            }
        )
        frame["ints"] = frame["ints"].astype(int)
        sanitized_frame = pd.DataFrame(
            {
                "non_numeric": list("abcd"),
                "floats": [123, 1e12, -1e12, 0],
                "ints": [123, 2147483647, -2147483648, 0],
            }
        )
        sanitized_frame["floats"] = sanitized_frame["floats"].astype(np.float32)
        sanitized_frame["ints"] = sanitized_frame["ints"].astype(np.int32)

        # Check sanitize_series
        for col in frame:
            assert all(sanitized_frame[col] == sanitize_series(frame[col]))

        # Check sanitize_dataframe
        assert all(sanitized_frame == sanitize_dataframe(frame))


class TestBaseFeatureProcessor:
    def test_mode(self):
        # Test notset
        fp = BaseFeatureProcessor(mode=None)
        assert fp.mode == "notset"
        fp = BaseFeatureProcessor(mode="notset")
        assert fp.mode == "notset"
        # Test classification
        fp = BaseFeatureProcessor(mode="classification")
        assert fp.mode == "classification"
        fp = BaseFeatureProcessor(mode="multiclass")
        assert fp.mode == "classification"
        # Test regression
        fp = BaseFeatureProcessor(mode="regression")
        assert fp.mode == "regression"
        # Test cases
        fp = BaseFeatureProcessor(mode="NOTSET")
        assert fp.mode == "notset"
        # Test error
        with pytest.raises(ValueError):
            # Should error when trying to set an unknown mode.
            BaseFeatureProcessor(mode="invalid_mode")

    @pytest.mark.parametrize("mode", ["classification"])
    def test_check_x(self, make_good_and_corrupt_classification):
        fp = BaseFeatureProcessor  # has not to be instantiated
        x_good, x_corrupt, _, _ = make_good_and_corrupt_classification

        # Check good input
        checked = fp._check_x(x_good)
        assert np.allclose(checked, x_good), "Good x data was handled unexpectedly."

        # Check corrupted input
        checked = fp._check_x(x_corrupt)
        assert np.allclose(checked, x_good), "Invalid data sanitization."
        assert all(isinstance(c, str) for c in checked.columns), "Non-string columns."

        # Check error when non-unique column names
        x_corrupt_columns = x_corrupt.columns
        x_corrupt.columns = [0, 0, *x_corrupt.columns[2:]]
        with pytest.raises(ValueError):
            # Should raise error when receiving non-unique column names.
            fp._check_x(x_corrupt)
        x_corrupt.columns = x_corrupt_columns  # reset since fixture has scope="class"

    @pytest.mark.parametrize("mode", ["classification"])
    def test_check_y(self, make_good_and_corrupt_classification):
        fp = BaseFeatureProcessor  # has not to be instantiated
        _, _, y_good, y_corrupt = make_good_and_corrupt_classification

        # Check good input
        checked = fp._check_y(y_good)
        assert np.allclose(checked, y_good), "Good y data was handled unexpectedly."

        # Check corrupted input
        checked = fp._check_y(y_corrupt)
        assert np.allclose(checked, y_good), "Invalid data sanitization."
        assert isinstance(checked.name, str), "Non-string name."

    @pytest.mark.parametrize("mode", ["classification"])
    def test_check_x_y(self, make_good_and_corrupt_classification):
        fp = BaseFeatureProcessor  # has not to be instantiated
        x, _, y, _ = make_good_and_corrupt_classification

        # Check integrity
        y_short = y.iloc[:-2]
        with pytest.raises(ValueError):
            # Should raise error when receiving x and y of unequal length.
            fp._check_x_y(x, y_short)

        # Test warning for non-matching index
        y_bad_index = deepcopy(y)
        y_bad_index.index = pd.RangeIndex(1, len(y) + 1)
        with pytest.warns(UserWarning):
            fp._check_x_y(x, y_bad_index)


@pytest.mark.usefixtures("make_rng")
class TestBaseFeatureExtractor:
    rng: np.random.Generator

    def test_setting_features(self):
        fe = BaseFeatureExtractor()
        # Test set_features
        initial_features = [f"feat_{i}" for i in range(10)]
        fe.set_features(initial_features)
        assert set(fe.features_) == set(initial_features)
        # Test add_features
        added_features = [f"feat_{i}" for i in range(10, 20)]
        fe.add_features(added_features)
        fe.add_features(added_features)  # this should have no effect
        assert set(fe.features_) == set(initial_features).union(added_features)
        # Test remove_features
        fe.remove_features(added_features)
        assert set(fe.features_) == set(initial_features)
        with pytest.raises(ValueError):
            # Should raise error when trying to remove features that don't exist.
            fe.remove_features(["non_existing"])

    @pytest.mark.parametrize("mode", ["classification", "multiclass", "regression"])
    def test_scoring(self, mode):
        # Init
        fe = BaseFeatureExtractor(mode=mode)
        fe._set_validation_model()
        size = 100
        if mode == "classification":
            y = pd.Series([0] * size)
            y.iloc[::2] = 1
        elif mode == "multiclass":
            y = pd.Series([0] * size)
            y.iloc[::3] = 1
            y.iloc[1::3] = 2
        else:
            y = pd.Series(range(size))
        x = pd.DataFrame({"1to1": y, "random": self.rng.geometric(0.5, size)})

        # Test _calc_feature_scores
        scores = x.apply(fe._calc_feature_scores, y=y, axis=0)  # type: ignore
        assert all(scores["1to1"] >= 0.9)
        assert all(scores["random"] < 0.9)

        # Test _init_feature_baseline_scores
        fe._init_feature_baseline_scores(x, y)
        assert all(np.array(fe._baseline_scores) == scores.max(1).values)

        # Test _update_feature_baseline_scores
        new_scores = pd.DataFrame({"new_score": [1.0] * len(scores)})
        fe._update_feature_baseline_scores(new_scores)
        assert all(np.array(fe._baseline_scores) == 1.0)

    def test_accept_feature(self):
        fe = BaseFeatureExtractor()
        fe._baseline_scores = [0.5, 0.5]
        assert fe.accept_feature([0.6, 0.6])
        assert fe.accept_feature([0.6, 0.1])
        assert fe.accept_feature([0.1, 0.6])
        assert not fe.accept_feature([0.1, 0.1])

    def test_select_scores(self):
        fe = BaseFeatureExtractor()
        fe._baseline_scores = [0.5, 0.5]
        scores = pd.DataFrame(
            {
                "good_1": [0.6, 0.6],
                "good_2": [0.6, 0.1],
                "good_3": [0.1, 0.6],
                "bad": [0.1, 0.1],
            }
        )
        accepted_scores = fe.select_scores(scores)
        assert set(accepted_scores) == set(scores.drop("bad", axis=1)), "Accepted bad."



#  Copyright (c) 2022 by Amplo.

from pathlib import Path

import numpy as np
import pandas as pd

from amplo.automl import IntervalAnalyser
from tests import create_data_frames, rmtree


class TestIntervalAnalyser:
    @classmethod
    def setup_class(cls):
        log_files = [f"log_{i}" for i in range(20)]
        index = pd.MultiIndex.from_product([log_files, range(100)])
        dfs = []
        labels = []
        for i in range(10):
            df1, df2 = create_data_frames(n_samples=100, n_features=5)
            dfs.extend([df1, df2])
            labels.extend([1] * 100)
            labels.extend([0] * 100)
        df = pd.concat(dfs, axis=0)
        df = df.set_index(index, drop=True)
        df = (df - df.min()) / (df.max() - df.min())
        df["target"] = labels
        cls.df = df

    def test_all(self):
        # Normal run
        ia = IntervalAnalyser(min_length=10, target="target")
        data_no_noise = ia.fit_transform(self.df)

        # Attribute tests
        assert ia.n_samples_ == 2000
        assert ia.n_columns_ == 5
        assert ia.n_folders_ == 2
        assert ia.n_files_ == 20

        # Functional tests
        assert ia.distributions_ is not None
        for log in ia.distributions_.index.get_level_values(0):
            dist = np.array(ia.distributions_[log].values)
            assert all(
                v < 0.8 for v in dist[: int(len(dist) / 2)]
            ), "Noise with high percentage of neighbors"
            assert all(
                v > 0.8 for v in dist[int(len(dist) / 2) :]
            ), "Information with low percentage of neighbors"

        # Data tests
        assert (
            len(self.df) == ia.n_samples_ and len(data_no_noise) * 2 == ia.n_samples_
        ), "Incorrect number of samples"
        assert (
            self.df.index.get_level_values(0).nunique() == ia.n_files_
        ), "Files skipped"
        assert len(self.df.keys()) == len(
            data_no_noise.keys()
        ), "Incorrect number of features"

    # TODO test correlation warning
    # TODO test data quality warning
    # TODO test multi-index error
    # TODO test index misalignment error
    # TODO test feature/sample length error
    # TODO test target != labels.name



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import json
import os
import re
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Iterable
from warnings import warn

import numpy as np
import pandas as pd
from requests import HTTPError

from amplo.api.storage import AzureBlobDataAPI
from amplo.utils.logging import get_root_logger

if TYPE_CHECKING:
    from logging import Logger

    from amplo.api.platform import AmploPlatformAPI

__all__ = [
    "boolean_input",
    "parse_json",
    "NpEncoder",
    "read_pandas",
    "get_file_metadata",
    "merge_logs",
]


FILE_READERS = {
    ".csv": pd.read_csv,
    ".json": pd.read_json,
    ".xml": pd.read_xml,
    ".feather": pd.read_feather,
    ".parquet": pd.read_parquet,
    ".stata": pd.read_stata,
    ".pickle": pd.read_pickle,
}


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, pd.Series):
            return obj.to_list()
        if isinstance(obj, pd.DataFrame):
            obj.to_json()
        return super(NpEncoder, self).default(obj)


def boolean_input(question: str) -> bool:
    x = input(question + " [y / n]")
    if x.lower() == "n" or x.lower() == "no":
        return False
    elif x.lower() == "y" or x.lower() == "yes":
        return True
    else:
        warn('Sorry, I did not understand. Please answer with "n" or "y"')
        return boolean_input(question)


def parse_json(json_string: str | dict) -> str | dict:
    if isinstance(json_string, dict):
        return json_string
    else:
        try:
            return json.loads(
                json_string.replace("'", '"')
                .replace("True", "true")
                .replace("False", "false")
                .replace("nan", "NaN")
                .replace("None", "null")
            )
        except json.decoder.JSONDecodeError:
            warn(f"Cannot validate, impassable JSON: {json_string}")
            return json_string


def read_pandas(path: str | Path) -> pd.DataFrame:
    """
    Wrapper for various read functions

    Returns
    -------
    pd.DataFrame
    """
    file_extension = Path(path).suffix
    if file_extension not in FILE_READERS:
        raise NotImplementedError(f"File format {file_extension} not supported.")
    else:
        reader = FILE_READERS[file_extension]
        return reader(path, low_memory=False)


def get_file_metadata(file_path: str | Path) -> dict[str, str | float]:
    """
    Get file metadata from given path.

    Parameters
    ----------
    file_path : str or Path
        File path.

    Returns
    -------
    dict of {str: str or float}
        File metadata.

    Raises
    ------
    FileNotFoundError
        When the path does not exist.
    IsADirectoryError
        When the path resolves a directory, not a file.
    """

    from amplo.utils import check_dtypes

    check_dtypes("file_path", file_path, (str, Path))

    file_path = Path(file_path)

    if not file_path.exists():
        raise FileNotFoundError(f"File does not exist: '{file_path}'")
    if not file_path.is_file():
        raise IsADirectoryError(f"Path is not a file: '{file_path}'")

    return {
        "file_name": str(file_path.name),
        "full_path": str(file_path.resolve()),
        # "creation_time": os.path.getctime(str(file_path)),
        "last_modified": os.path.getmtime(str(file_path)),
    }


def _get_api_clients(
    azure: tuple[str, str] | bool = False,
    platform: tuple[str, str] | bool | None = None,
):
    """
    Gathers the api clients for merge_logs
    """
    from amplo.api.platform import AmploPlatformAPI

    if not azure:
        blob_api = None
    else:
        azure = azure if not isinstance(azure, bool) else tuple()
        blob_api = AzureBlobDataAPI.from_os_env(*azure)

    # Mirror azure parameter when platform is not set
    if platform is None:
        platform = bool(azure)
    # Get amplo platform client
    if not platform:
        platform_api = None
    else:
        platform = platform if not isinstance(platform, bool) else tuple()
        platform_api = AmploPlatformAPI.from_os_env(*platform)
    return blob_api, platform_api


def _get_folders(
    parent_folder: str | Path,
    blob_api: AzureBlobDataAPI | None = None,
    more_folders: list[str | Path] | None = None,
) -> list[Path]:
    """
    Lists folders for merge_logs
    """
    if not blob_api:
        if not Path(parent_folder).exists():
            raise ValueError(f"{parent_folder} directory does not exist.")
        folders = [
            folder for folder in Path(parent_folder).iterdir() if folder.is_dir()
        ]
    else:
        folders = blob_api.ls_folders(parent_folder)

    # Add more_folders
    if more_folders:
        folders += more_folders

    # return [Path(f) for f in folders]
    return [Path(f) for i, f in enumerate(folders) if i < 3]


def _read_files_in_folders(
    folders: Iterable[str | Path],
    target: str,
    blob_api: AzureBlobDataAPI | None = None,
    logger: Logger | None = None,
) -> tuple[list[str], pd.DataFrame, dict[str, dict]]:
    """
    Use pandas to read all non-hidden and non-empty files into a DataFrame.

    Parameters
    ----------
    folders : iterable of (str or Path)
        Directory names.
    target : str
        Target column & directory name
    blob_api : AzureBlobDataAPI or None, optional, default: None
        If None, tries to read data from local folder, else from Azure
    logger : Logger or None, optional, default: None
        When provided, will log progress every 90 seconds.

    Returns
    -------
    filenames : list of str
    data : pd.DataFrame
        All files of the folders merged into one multi-indexed DataFrame.
    metadata : list of dict of {str : str or float}

    Warnings
    --------
    UserWarning
        When any directory is empty, or has no supported file type.
    """

    # Map folders to pathlib.Path object
    folders = [Path(f) for f in folders]

    # Initialize
    file_names, data, metadata = [], [], {}
    last_time_logged = time()
    for folder_count, folder in enumerate(sorted(folders)):

        # List all files
        if blob_api:
            files = list(map(Path, blob_api.ls_files(folder)))
        else:
            files = [f for f in folder.iterdir() if f.is_file()]

        # Remove hidden files
        hidden_files = [f for f in files if re.match(r"^\..*", f.name)]
        files = list(set(files) - set(hidden_files))

        # Remove unsupported file types
        unsupported_files = [f for f in files if f.suffix not in FILE_READERS]
        files = list(set(files) - set(unsupported_files))

        # Remove empty files
        if blob_api:
            empty_files = [f for f in files if blob_api.get_size(f) == 0]
        else:
            empty_files = [f for f in files if f.stat().st_size == 0]
        files = list(set(files) - set(empty_files))

        # Sanity check
        if not files:
            warn(f"Directory is empty and thus skipped: '{folder}'")
            continue

        # Read files
        for file_ in sorted(files):

            # read_pandas() may raise an EmptyDataError when the file has no content.
            # The try...except catches such errors and warns the user instead.
            try:
                if blob_api:
                    datum = blob_api.read_pandas(file_)
                    metadatum = blob_api.get_metadata(file_)
                else:
                    datum = read_pandas(file_)
                    metadatum = get_file_metadata(file_)
            except pd.errors.EmptyDataError:
                warn(f"Empty file detected and thus skipped: '{file_}'")
                continue

            # Convert to dataframe
            if isinstance(datum, pd.Series):
                datum = datum.to_frame()

            # Set multi-index
            datum = datum.set_index(
                pd.MultiIndex.from_product(
                    [[str(file_)], datum.index.values], names=["log", "index"]
                )
            )

            # Add target
            datum[target] = (folder.name == target) * 1

            # Append
            file_names.append(str(file_))
            data.append(datum)
            metadata[str(file_)] = metadatum

        if logger and time() - last_time_logged > 90:
            last_time_logged = time()
            logger.info(f".. progress: {folder_count / len(folders) * 100:.1f} %")

    # Concatenate data
    data = pd.concat(data, axis=0)

    # Validate data
    if target not in data:
        raise ValueError("Target not in data.")
    if data[target].nunique() != 2:
        raise ValueError(f"Number of unique labels is {data[target].nunique()} != 2.")

    return file_names, data, metadata


def _map_datalogs_to_file_names(
    file_names: list[str],
    platform_api: AmploPlatformAPI | None = None,
    logger: Logger | None = None,
) -> dict:
    """
    Get datalogs for every filename.

    Parameters
    ----------
    file_names : list of str
        Files names to get datalogs from - if available.
    platform_api : AmploPlatformAPI or None, optional, default: None
        API to get datlogs from.
    logger : Logger or None, optional, default: None
        When provided, will log progress every 90 seconds.

    Returns
    -------
    list of dict
        Datalogs for every filename.
    """

    if not platform_api:
        return {}

    # It is assumed that the 6th and 5th path position of the (first) filename contains
    # the team and machine name, respectively, if you count from right to left.
    # E.g., "Team/Machine/data/Category/Issue/log_file.csv"

    # Remove path prefixes, otherwise datalogs will not be found
    file_names = ["/".join(str(fname).split("/")[-6:]) for fname in file_names]
    # Extract team and machine
    try:
        team, machine = file_names[0].split("/")[-6:-4]
    except IndexError:
        warn(f"Got an empty list of file names")
        return {}

    # Get datalog for each filename
    datalogs = {}
    last_time_logged = time()
    for file_count, fname in enumerate(file_names):
        try:
            datalog = platform_api.get_datalog(team, machine, fname)
        except HTTPError:
            # No matching datalog found. Do still append it to preserve the order.
            datalog = {}

        datalogs[fname] = datalog

        if logger and time() - last_time_logged > 90:
            last_time_logged = time()
            logger.info(f".. progress: {file_count / len(file_names) * 100:.1f} %")

    return datalogs


def _mask_intervals(datalogs: dict, data: pd.DataFrame) -> pd.DataFrame:
    """
    Masks the data with the intervals given by the datalogs.

    Parameters
    ----------
    datalogs : list of dict
        Datalogs dictionary that should contain the keys 'selected' and 'datetime_col'.
    data : pd.DataFrame
        Data for splitting.

    Returns
    -------
    data_out : pd.DataFrame
        Selected data.

    Warnings
    --------
    UserWarning
        When no valid match for the start or stop time of the data interval was found,
        i.e. when the time difference is more than 1 second.
    """
    for filename in data.index.get_level_values("log").unique():
        # Get intervals and timestamp column from datalog
        datalog = datalogs.get(filename, {})
        intervals = datalog.get("selected", [])
        ts_col = datalog.get("datetime_col", "")

        # Validate
        if not intervals or not ts_col:
            continue
        elif ts_col not in data:
            warn(f"Cannot select intervals as the column '{ts_col}' is not present.")
            continue

        # Convert ts_col
        if not pd.api.types.is_numeric_dtype(data[ts_col]):
            data[ts_col] = (
                pd.to_datetime(data[ts_col], errors="coerce").astype(int) / 10**9
            )

        # Extract intervals
        drop_mask = True
        for interval in intervals:
            ts_first, ts_last = interval
            drop_mask = (
                (data[ts_col] < ts_first) | (data[ts_col] > ts_last)
            ) & drop_mask
        data = data.drop(data.loc[(filename, drop_mask), :].index)

    return data


def merge_logs(
    parent: str | Path,
    target: str,
    more_folders: list[str | Path] | None = None,
    azure: tuple[str, str] | bool = False,
    platform: tuple[str, str] | bool | None = None,
) -> tuple[pd.DataFrame, dict[str, dict]]:
    """
    Combine log files of all subdirectories into a multi-indexed DataFrame.

    The function can handle logs from a local directory as well as data coming from an
    Azure blob storage. For the latter case it is furthermore capable to select
    intervals using Amplo's datalogs.

    Notes
    -----
    Make sure that each protocol is located in a subdirectory whose name represents the
    respective label.

    An exemplary directory structure of ``parent_folder``:
        ``
        parent_folder
        ├─ Label_1
        │   ├─ Log_1.*
        │   └─ Log_2.*
        ├─ Label_2
        │   └─ Log_3.*
        └─ ...
        ``

    Parameters
    ----------
    parent_folder : str or Path
        Directory that contains subdirectories with tabular data files.
    target : str
        The target folder.
    more_folders : list of str or Path, optional
        Additional folder names with tabular data files to append.
    azure : (str, str) or bool, default: False
        Use this parameter to indicate that data is in Azure blob storage.
        If False, it is assumed that data origins from local directory.
        If True, the AzureBlobDataAPI is initialized with default OS env variables.
        Otherwise, it will use the tuple to initialize the api.
    platform : (str, str) or bool or None, default: None
        Use this parameter for selecting data according to Amplo's datalogs.
        If None, its value is set to bool(azure).
        If False, no datalogs information will be used.
        If True, the AmploPlatformAPI is initialized with default OS env variables.
        Otherwise, it will use the tuple to initialize the api.
    verbose : int, default = 1


    Returns
    -------
    data : pd.DataFrame
        All files of the folders merged into one multi-indexed DataFrame.
        Multi-index names are 'log' and 'index'.
    metadata : dict of {int : dict of {str : str or float}}
        Metadata of merged data.
    """
    from amplo.utils import check_dtypes

    logger = get_root_logger()
    check_dtypes(
        ("parent_folder", parent, (str, Path)),
        ("target_col", target, str),
        ("more_folders", more_folders, (type(None), list)),
    )

    # Get azure blob client
    blob_api, platform_api = _get_api_clients(azure, platform)

    # Get child folders
    folders = _get_folders(parent, blob_api)
    if target not in [f.name for f in folders]:
        raise ValueError(f"Target {target} not present in folders.")
    logger.info(f"Found {len(folders)} folders.")

    # Pandas read files
    fnames, data, metadata = _read_files_in_folders(folders, target, blob_api, logger)
    logger.info(f"Found {len(fnames)} files.")

    # Masking data
    if platform_api:
        logger.info("Reading datalogs from platform")
        datalogs = _map_datalogs_to_file_names(fnames, platform_api, logger)
        if datalogs:
            logger.info("Masking intervals from datalogs")
            data = _mask_intervals(datalogs, data)

    return data, metadata



#  Copyright (c) 2022 by Amplo.

from abc import ABCMeta

from amplo.base import BasePredictor, LoggingMixin

__all__ = ["BaseRegressor"]


class BaseRegressor(BasePredictor, LoggingMixin, metaclass=ABCMeta):
    _estimator_type = "regressor"

    def __init__(self, model, verbose=0):
        BasePredictor.__init__(self)
        LoggingMixin.__init__(self, verbose=verbose)

        self.model = model

    def _fit(self, x, y=None, **fit_params):
        self.model.fit(x, y)

    def _predict(self, x, y=None, **predict_params):
        return self.model.predict(x, **predict_params).reshape(-1)

    def score(self, x, y):
        return self.model.score(x, y)

    def _get_model_params(self, deep=True):
        """Gets JSON serializable model parameters only."""
        return self.model.get_params(deep=deep)

    def get_params(self, deep=True):
        params = super().get_params(deep=deep)
        model_params = self._get_model_params(deep=deep)

        return {**model_params, **params}

    def set_params(self, **params):
        # Set class params
        valid_class_params = super().get_params(deep=True)
        class_params = {key: params[key] for key in params if key in valid_class_params}
        super().set_params(**class_params)
        # Set model params
        model_params = {key: params[key] for key in params if key not in class_params}
        self.model.set_params(**model_params)
        return self



#  Copyright (c) 2022 by Amplo.

import json
from datetime import datetime, timedelta

import joblib
import numpy as np
import pandas as pd
import pytest
from sklearn.metrics import log_loss, r2_score

from amplo import Pipeline
from amplo.utils.testing import make_interval_data
from tests import rmtree


@pytest.mark.usefixtures("make_mode")
class TestPipelineByMode:
    target: str
    mode: str
    data: pd.DataFrame

    def test_mode_detector(self):
        # Numerical test
        pipeline = Pipeline(no_dirs=True, target=self.target)
        pipeline._mode_detector(self.data)
        assert pipeline.mode == self.mode

        # Categorical test
        if self.mode == "classification":
            # Convert to categorical
            data = self.data.copy()
            data[self.target] = [f"Class_{v}" for v in data[self.target].values]
            # Detect mode
            pipeline = Pipeline(no_dirs=True, target=self.target)
            pipeline._mode_detector(self.data)
            assert pipeline.mode == self.mode

    def test_dir_and_settings(self):
        pipeline = Pipeline(
            target=self.target,
            mode=self.mode,
            objective="r2" if self.mode == "regression" else "neg_log_loss",
            n_grid_searches=0,
            extract_features=False,
        )
        pipeline.fit(self.data)

        if self.mode == "classification":
            # Pipeline Prediction
            prediction = pipeline.predict_proba(self.data)
            assert log_loss(self.data[self.target], prediction) > -1

        elif self.mode == "regression":
            # Pipeline Prediction
            prediction = pipeline.predict(self.data)
            assert len(prediction.shape) == 1
            assert r2_score(self.data[self.target], prediction) > 0.75

        else:
            raise ValueError(f"Invalid mode {self.mode}")

        # Settings prediction
        settings = json.load(open("Auto_ML/Settings.json", "r"))
        model = joblib.load("Auto_ML/Model.joblib")
        p = Pipeline(no_dirs=True)
        p.load_settings(settings)
        p.load_model(model)
        if self.mode == "classification":
            assert np.allclose(p.predict_proba(self.data), prediction)
        else:
            assert np.allclose(p.predict(self.data), prediction)

        # Test loading function
        p.load()
        if self.mode == "classification":
            assert np.allclose(p.predict_proba(self.data), prediction)
        else:
            assert np.allclose(p.predict(self.data), prediction)

        # Check settings
        assert p.best_model_
        assert p.best_model_str_
        assert p.best_params_
        assert p.best_feature_set_
        assert p.best_score_
        assert "pipeline" in settings
        assert "version" in settings
        assert "model" in settings
        assert "amplo_version" in settings
        assert "params" in settings
        assert "features" in settings
        assert "validation" in settings
        assert "feature_processing" in settings
        assert "data_processing" in settings


class TestPipeline:
    def test_drift(self):
        # todo connect with logger
        pytest.skip("This test is not yet implemented")

    def test_all_models(self):
        # TODO
        pytest.skip("This test is not yet implemented")

    def test_interval_analyzer(self):
        """
        Use interval analyzer when a folder with logs is presented to pipeline.
        """
        # Create dummy data
        data_dir = "./test_data"
        make_interval_data(directory=data_dir, target="target")

        # Pipeline
        pipeline = Pipeline(
            n_grid_searches=0, extract_features=False, interval_analyse=True
        )
        pipeline.fit(data_dir, target="DummyLabel_0")

        # Check if IntervalAnalyser is fitted
        assert (
            pipeline.interval_analyser and pipeline.interval_analyser.is_fitted_
        ), "IntervalAnalyser was not fitted"

        # Remove dummy data
        rmtree(data_dir, must_exist=True)

    def test_with_time(self):
        x = np.vstack(
            (  # type: ignore
                [datetime(2020, 1, 1) + timedelta(days=i) for i in range(12)],
                np.linspace(0, 10, 12),
            )
        ).T
        y = np.linspace(0, 10, 12)
        pipeline = Pipeline(n_grid_searches=0, extract_features=False)
        pipeline.fit(x, y)

        # Do the same, but with temporal
        multi_index = pd.MultiIndex.from_product(
            [[1, 2, 3], [1, 2, 3, 4]], names=["log", "index"]
        )
        df = pd.DataFrame(x, index=multi_index)
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(df, y)

    def test_backwards_compatibility(self):
        # Set up pipeline
        pipeline = Pipeline(main_dir="tests/files/")
        pipeline.load()

        # Predict
        np.random.seed(100)
        df = pd.DataFrame(
            {
                "output_current": np.random.normal(0, 1, size=100),
                "radiator_temp": np.random.normal(0, 1, size=100),
            }
        )
        yp = pipeline.predict_proba(df)
        assert np.allclose(yp, np.array([[0.16389499, 0.83610501]]))



#  Copyright (c) 2022 by Amplo.

import sys
from collections import deque
from collections.abc import Mapping, Set
from numbers import Number

import pandas as pd

__all__ = ["getsize"]

ZERO_DEPTH_BASES = (str, bytes, Number, range, bytearray)


def getsize(obj_0):
    # Credits: https://stackoverflow.com/questions/449560/how-do-i-determine-the-size-of-an-object-in-python  # noqa: E501
    """
    Recursively iterate to sum size of object & members.

    Parameters
    ----------
    obj_0 : Any
        Object whose size is estimated recursively
    """
    _seen_ids = set()

    def inner(obj):
        obj_id = id(obj)
        if obj_id in _seen_ids:
            return 0
        _seen_ids.add(obj_id)

        try:
            if isinstance(obj, (pd.Series, pd.DataFrame)):
                return obj.memory_usage(deep=True)

            size = sys.getsizeof(obj)
            if isinstance(obj, ZERO_DEPTH_BASES):
                pass  # bypass remaining control flow and return
            elif isinstance(obj, (tuple, list, Set, deque)):
                size += sum(inner(i) for i in obj)
            elif isinstance(obj, Mapping) or hasattr(obj, "items"):
                size += sum(inner(k) + inner(v) for k, v in getattr(obj, "items")())
            # Check for custom object instances - may subclass above too
            if hasattr(obj, "__dict__"):
                size += inner(vars(obj))
            if hasattr(obj, "__slots__"):  # can have __slots__ with __dict__
                size += sum(
                    inner(getattr(obj, s)) for s in obj.__slots__ if hasattr(obj, s)
                )
            return size
        except:
            return 0

    return inner(obj_0)



#  Copyright (c) 2022 by Amplo.

import subprocess

import setuptools

if __name__ == "__main__":
    version = (
        subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
        .stdout.decode("utf-8")
        .strip()
    )
    if "." not in version:
        raise ValueError(f"Invalid version: {version}")

    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    with open("requirements.txt") as f:
        required = f.read().splitlines()

    setuptools.setup(
        name="Amplo",
        version=version,
        author="Amplo AG",
        author_email="info@amplo.ch",
        description="Fully automated end to end machine learning pipeline",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/nielsuit227/AutoML",
        packages=setuptools.find_packages(),
        package_data={"Amplo": ["VERSION"]},
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: OS Independent",
        ],
        python_requires=">=3.9",
        install_requires=required,
    )



#  Copyright (c) 2022 by Amplo.

from amplo.regression._base import BaseRegressor
from amplo.regression.catboost import CatBoostRegressor
from amplo.regression.lgbm import LGBMRegressor
from amplo.regression.partial_boosting import PartialBoostingRegressor
from amplo.regression.stacking import StackingRegressor
from amplo.regression.xgb import XGBRegressor



#  Copyright (c) 2022 by Amplo.
from __future__ import annotations

import copy
import re
import time
from datetime import datetime
from warnings import warn

import numpy as np
import optuna
import pandas as pd
from sklearn.metrics import get_scorer
from sklearn.model_selection import BaseCrossValidator, StratifiedKFold, cross_val_score

from amplo.base.objects import BasePredictor, LoggingMixin

__all__ = ["OptunaGridSearch"]


MIN_REGULARIZATION = 1e-6
MAX_REGULARIZATION = 1e2
MIN_BOOSTERS = 100
MAX_BOOSTERS = 2500
MIN_DEPTH = 1
MAX_DEPTH = 10
MIN_SPLIT = 0.1
MAX_SPLIT = 1.0
MIN_LEAF_SIZE = 1
MAX_LEAF_SIZE = 10_000

# Define min-max-function
def minimax(min_, value, max_):
    return max(min_, min(value, max_))


def max_depth(samples: int) -> int:
    return minimax(MIN_DEPTH, int(np.log2(samples)), MAX_DEPTH)


def max_leaf_size(samples: int) -> int:
    return minimax(MIN_LEAF_SIZE, int(samples / 10), MAX_LEAF_SIZE)


def warn_at_extreme(params: dict[str, int | float | str], samples: int):
    for param, value in params.items():
        if param in ["depth", "max_depth"]:
            if value == MIN_DEPTH or value == max_depth(samples):
                warn(
                    f"Parameter {param} (={value}) at the edge of search space: ({MIN_DEPTH}, {max_depth(samples)})"
                )
        if params in [
            "n_estimators",
            "alpha",
            "C",
            "l2_leaf_reg",
            "learning_rate",
            "eta",
            "gamma",
            "l2_regularization",
            "skip_drop",
            "rate_drop",
            "min_sum_hessian_in_leaf",
            "reg_alpha",
            "reg_lambda",
            "lambda_l1",
            "lambda_l1",
        ]:
            if value == MIN_REGULARIZATION or value == MAX_REGULARIZATION:
                warn(
                    f"Parameter {param} (={value}) at the edge of search space: ({MIN_REGULARIZATION}, {MAX_REGULARIZATION})"
                )
        if param in ["max_iter", "n_estimators"]:
            if value == MIN_BOOSTERS or value == MAX_BOOSTERS:
                warn(
                    f"Parameter {param} (={value}) at the edge of search space: ({MIN_BOOSTERS}, {MAX_BOOSTERS})"
                )
        if param in [
            "max_samples",
            "max_features",
            "colsample_bytree",
            "feature_fraction",
            "bagging_fraction",
            "subsample",
        ]:
            if value == MIN_SPLIT or value == MAX_SPLIT:
                warn(
                    f"Parameter {param} (={value}) at the edge of search space: ({MIN_SPLIT}, {MAX_SPLIT})"
                )
        if param in ["leaf_size", "min_data_in_leaf", "min_samples_leaf"]:
            if value == MIN_LEAF_SIZE or value == max_leaf_size(samples):
                warn(
                    f"Parameter {param} (={value}) at the edge of search space: ({MIN_LEAF_SIZE}, {max_leaf_size(samples)})"
                )


class OptunaGridSearch(LoggingMixin):
    """
    Wrapper for ``optuna`` grid search.

    Takes any model supported by `Amplo.AutoML.Modelling` whose parameter
    search space is predefined for each model.
    Optimal choice [1]

    [1] https://arxiv.org/pdf/2201.06433.pdf

    Parameters
    ----------
    model : Amplo.AutoML.Modeller.ModelType
        Model object to optimize.
    n_trials : int
        Limit the number of trials/candidates to search.
    timeout : int
        Limit the time for optimization.
    cv : sklearn.model_selection.BaseCrossValidator
        Cross validation object.
    scoring : str or sklearn.metrics._scorer._BaseScorer, default = neg_log_loss
        A valid string for `sklearn.metrics.get_scorer`.
    verbose : int
        Verbose logging.
    """

    def __init__(
        self,
        model: BasePredictor,
        target: str,
        n_trials: int = 250,
        timeout: int = -1,
        cv: BaseCrossValidator = StratifiedKFold(n_splits=10),
        scoring: str = "neg_log_loss",
        verbose: int = 0,
    ):
        super().__init__(verbose=verbose)

        # Input tests
        if hasattr(model, "is_fitted") and model.is_fitted:
            warn(
                "The model is already fitted but Amplo's grid search will re-fit.",
                UserWarning,
            )
        scoring = get_scorer(scoring)  # type: ignore

        # Set class attributes
        self.target = target
        self.model = model
        self.n_trials = n_trials
        self.timeout = timeout
        self.cv = cv
        self.scoring = get_scorer(scoring)

        # Set attributes
        self.binary_ = None
        self.samples_ = None
        self.trial_count_ = -1

        # Model specific settings
        if (
            type(self.model).__name__ == "LinearRegression"
            or type(self.model).__name__ == "LogisticRegression"
        ):
            self.n_trials = 1

    def _get_hyper_params(
        self, trial: optuna.Trial
    ) -> dict[str, None | bool | int | float | str]:
        """Get model specific hyper parameter values, indicating predefined
        search areas to optimize.

        parameters
        ----------
        trial : optuna.Trial
        """
        assert self.samples_ and self.binary_ and self.x
        # Extract model name & type
        model_name = type(self.model).__name__
        model_type = re.split(r"Regressor|Classifier", model_name)[0]

        # Determine whether it's classification or regression
        is_regression = bool(re.match(r".*(Regression|Regressor|SVR)", model_name))
        is_classification = bool(
            re.match(r".*(Classification|Classifier|SVC)", model_name)
        )
        if not (is_regression or is_classification):
            raise ValueError("Could not determine mode (regression or classification).")

        # Find matching model and return its parameter values
        if model_name == "LinearRegression" or model_name == "LogisticRegression":
            return {}

        elif model_name == "Lasso" or "Ridge" in model_name:
            return dict(
                alpha=trial.suggest_float(
                    "alpha", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                )
            )

        elif model_name in ("SVR", "SVC"):
            return dict(
                gamma=trial.suggest_categorical(
                    "gamma", ["scale", "auto", 0.001, 0.01, 0.1, 0.5, 1]
                ),
                C=trial.suggest_float(
                    "C", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
            )

        elif model_type == "KNeighbors":
            return dict(
                n_neighbors=trial.suggest_int(
                    "n_neighbors", 5, minimax(5, self.samples_ // 10, 50)
                ),
                weights=trial.suggest_categorical("weights", ["uniform", "distance"]),
                leaf_size=trial.suggest_int(
                    "leaf_size", MIN_LEAF_SIZE, max_leaf_size(self.samples_), log=True
                ),
            )

        elif model_type == "MLP":
            raise NotImplementedError("MLP is not supported")

        elif model_type == "SGD":
            params = dict(
                loss=trial.suggest_categorical(
                    "loss",
                    [
                        "squared_loss",
                        "huber",
                        "epsilon_insensitive",
                        "squared_epsilon_insensitive",
                    ],
                ),
                penalty=trial.suggest_categorical(
                    "penalty", ["l2", "l1", "elasticnet"]
                ),
                alpha=trial.suggest_float(
                    "alpha", MIN_REGULARIZATION, MAX_REGULARIZATION
                ),
                max_iter=trial.suggest_int(
                    "max_iter", MIN_BOOSTERS, MAX_BOOSTERS, log=True
                ),
            )
            if is_classification:
                params.update(
                    loss=trial.suggest_categorical(
                        "loss",
                        ["hinge", "log", "modified_huber", "squared_hinge"],
                    )
                )
            return params

        elif model_type == "DecisionTree":
            params = dict(
                criterion=trial.suggest_categorical(
                    "criterion",
                    ["squared_error", "friedman_mse", "absolute_error", "poisson"],
                ),
                max_depth=trial.suggest_int(
                    "max_depth", MIN_DEPTH, max_depth(self.samples_)
                ),
            )
            if is_classification:
                params.update(
                    criterion=trial.suggest_categorical(
                        "criterion", ["gini", "entropy"]
                    )
                )
            return params

        elif model_type == "AdaBoost":
            params = dict(
                n_estimators=trial.suggest_int(
                    "n_estimators", MIN_BOOSTERS, MAX_BOOSTERS, log=True
                ),
                loss=trial.suggest_categorical(
                    "loss", ["linear", "square", "exponential"]
                ),
                learning_rate=trial.suggest_float(
                    "learning_rate", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
            )
            if is_classification:
                params.pop("loss", None)
            return params

        elif model_type == "Bagging":
            return dict(
                n_estimators=trial.suggest_int(
                    "n_estimators", MIN_BOOSTERS, MAX_BOOSTERS, log=True
                ),
                max_samples=trial.suggest_float("max_samples", MIN_SPLIT, MAX_SPLIT),
                max_features=trial.suggest_float("max_features", MIN_SPLIT, MAX_SPLIT),
            )

        elif model_type == "CatBoost":
            params = dict(
                n_estimators=trial.suggest_int(
                    "n_estimators", MIN_BOOSTERS, MAX_BOOSTERS, log=True
                ),
                loss_function=trial.suggest_categorical(
                    "loss_function", ["MAE", "RMSE"]
                ),
                learning_rate=trial.suggest_float(
                    "learning_rate", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
                l2_leaf_reg=trial.suggest_float(
                    "l2_leaf_reg", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
                depth=trial.suggest_int("depth", MIN_DEPTH, max_depth(self.samples_)),
                min_data_in_leaf=trial.suggest_int(
                    "min_data_in_leaf",
                    MIN_LEAF_SIZE,
                    max_leaf_size(self.samples_),
                    log=True,
                ),
                grow_policy=trial.suggest_categorical(
                    "grow_policy",
                    ["SymmetricTree", "Depthwise", "Lossguide"],
                ),
                od_pval=1e-5,
                verbose=0,
            )
            if is_classification:
                params.update(
                    loss_function=trial.suggest_categorical(
                        "loss_function",
                        ["Logloss" if self.binary else "MultiClass"],
                    )
                )
            return params

        elif model_type == "GradientBoosting":
            params = dict(
                loss=trial.suggest_categorical("loss", ["ls", "lad", "huber"]),
                learning_rate=trial.suggest_float(
                    "learning_rate", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
                max_depth=trial.suggest_int(
                    "max_depth", MIN_DEPTH, max_depth(self.samples_)
                ),
                n_estimators=trial.suggest_int(
                    "n_estimators", MIN_BOOSTERS, MAX_BOOSTERS, log=True
                ),
                min_samples_leaf=trial.suggest_int(
                    "min_samples_leaf",
                    MIN_LEAF_SIZE,
                    max_leaf_size(self.samples_),
                    log=True,
                ),
                max_features=trial.suggest_float(
                    "max_features", MIN_SPLIT, MAX_SPLIT, log=True
                ),
                subsample=trial.suggest_float(
                    "subsample", MIN_SPLIT, MAX_SPLIT, log=True
                ),
            )
            if is_classification:
                params.update(
                    loss=trial.suggest_categorical(
                        "categorical", ["deviance", "exponential"]
                    )
                )
            return params

        elif model_type == "HistGradientBoosting":
            params = dict(
                loss=trial.suggest_categorical(
                    "loss", ["least_squares", "least_absolute_deviation"]
                ),
                learning_rate=trial.suggest_float(
                    "learning_rate", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
                max_iter=trial.suggest_int(
                    "max_iter", MIN_BOOSTERS, MAX_BOOSTERS, log=True
                ),
                max_leaf_nodes=trial.suggest_int("max_leaf_nodes", 30, 150, log=True),
                max_depth=trial.suggest_int(
                    "max_depth", MIN_DEPTH, max_depth(self.samples_)
                ),
                min_samples_leaf=trial.suggest_int(
                    "min_samples_leaf",
                    MIN_LEAF_SIZE,
                    max_leaf_size(self.samples_),
                    log=True,
                ),
                l2_regularization=trial.suggest_float(
                    "l2_regularization",
                    MIN_REGULARIZATION,
                    MAX_REGULARIZATION,
                    log=True,
                ),
                max_bins=trial.suggest_int("max_bins", 100, 255),
                early_stopping=True,
            )
            if is_classification:
                params.pop("loss")
            return params

        elif model_type == "RandomForest":
            params = dict(
                n_estimators=trial.suggest_int(
                    "n_estimators", MIN_BOOSTERS, MAX_BOOSTERS
                ),
                criterion=trial.suggest_categorical(
                    "criterion", ["squared_error", "absolute_error"]
                ),
                max_depth=trial.suggest_int(
                    "max_depth", MIN_DEPTH, max_depth(self.samples_)
                ),
                max_features=trial.suggest_categorical(
                    "max_features", ["log2", "sqrt"]
                ),
                min_samples_split=trial.suggest_int("min_samples_split", 2, 50),
                min_samples_leaf=trial.suggest_int(
                    "min_samples_leaf",
                    MIN_LEAF_SIZE,
                    max_leaf_size(self.samples_),
                    log=True,
                ),
                bootstrap=trial.suggest_categorical("bootstrap", [True, False]),
            )
            if is_classification:
                params.update(
                    criterion=trial.suggest_categorical(
                        "criterion", ["gini", "entropy"]
                    )
                )
            return params

        elif model_type == "XGB":
            params = dict(
                objective="reg:squarederror",
                eval_metric="rmse",
                booster=trial.suggest_categorical(
                    "booster", ["gbtree", "gblinear", "dart"]
                ),
                alpha=trial.suggest_float(
                    "alpha", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
                learning_rate=trial.suggest_float(
                    "learning_rate", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                ),
            )
            if is_classification:
                params.update(
                    eval_metric="logloss",
                )
            if params.get("booster") == "gbtree":
                params.update(
                    max_depth=trial.suggest_int(
                        "max_depth",
                        MIN_DEPTH,
                        max_depth(self.samples_),
                    ),
                    eta=trial.suggest_float(
                        "eta", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    gamma=trial.suggest_float(
                        "gamma", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    grow_policy=trial.suggest_categorical(
                        "grow_policy", ["depthwise", "lossguide"]
                    ),
                )
            elif params.get("booster") == "dart":
                params.update(
                    max_depth=trial.suggest_int(
                        "max_depth",
                        MIN_DEPTH,
                        max_depth(self.samples_),
                    ),
                    eta=trial.suggest_float(
                        "eta", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    gamma=trial.suggest_float(
                        "gamma", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    grow_policy=trial.suggest_categorical(
                        "grow_policy", ["depthwise", "lossguide"]
                    ),
                    sample_type=trial.suggest_categorical(
                        "sample_type", ["uniform", "weighted"]
                    ),
                    normalize_type=trial.suggest_categorical(
                        "normalize_type", ["tree", "forest"]
                    ),
                    rate_drop=trial.suggest_float(
                        "rate_drop", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    skip_drop=trial.suggest_float(
                        "skip_drop", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                )
            return params

        elif model_type == "LGBM":
            if is_regression:
                return dict(
                    num_leaves=trial.suggest_int("num_leaves", 10, 150),
                    min_data_in_leaf=trial.suggest_int(
                        "min_data_in_leaf",
                        MIN_LEAF_SIZE,
                        max_leaf_size(self.samples_),
                        log=True,
                    ),
                    min_sum_hessian_in_leaf=trial.suggest_float(
                        "min_sum_hessian_in_leaf",
                        MIN_REGULARIZATION,
                        MAX_REGULARIZATION,
                        log=True,
                    ),
                    colsample_bytree=trial.suggest_float(
                        "colsample_bytree", MIN_SPLIT, MAX_SPLIT, log=True
                    ),
                    reg_alpha=trial.suggest_float(
                        "reg_alpha", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    reg_lambda=trial.suggest_float(
                        "reg_lambda", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    verbosity=-1,
                )
            else:  # is_classification
                return dict(
                    objective="binary" if self.binary else "multiclass",
                    metric=trial.suggest_categorical(
                        "metric",
                        ["binary_error", "auc", "average_precision", "binary_logloss"]
                        if self.binary
                        else ["multi_error", "multi_logloss", "auc_mu"],
                    ),
                    boosting_type="gbdt",
                    lambda_l1=trial.suggest_float(
                        "lambda_l1", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    lambda_l2=trial.suggest_float(
                        "lambda_l2", MIN_REGULARIZATION, MAX_REGULARIZATION, log=True
                    ),
                    num_leaves=trial.suggest_int("num_leaves", 10, 5000),
                    max_depth=trial.suggest_int(
                        "max_depth", MIN_DEPTH, max_depth(self.samples_)
                    ),
                    min_data_in_leaf=trial.suggest_int(
                        "min_data_in_leaf",
                        MIN_LEAF_SIZE,
                        max_leaf_size(self.samples_),
                        log=True,
                    ),
                    min_gain_to_split=trial.suggest_float("min_gain_to_split", 0, 5),
                    feature_fraction=trial.suggest_float(
                        "feature_fraction", MIN_SPLIT, MAX_SPLIT, log=True
                    ),
                    bagging_fraction=trial.suggest_float(
                        "bagging_fraction", MIN_SPLIT, MAX_SPLIT, log=True
                    ),
                    bagging_freq=trial.suggest_int("bagging_freq", 1, 7),
                    verbosity=-1,
                )

        # Raise error if no match was found
        raise NotImplementedError(
            "Hyper parameter tuning not implemented for {}".format(model_name)
        )

    def fit(self, data: pd.DataFrame) -> pd.DataFrame:
        assert self.target in data
        self.y = data[self.target]
        self.x = data.drop(self.target, axis=1)

        # Set mode
        self.binary = self.y.nunique() == 2
        self.samples_ = len(self.y)

        # Set up study
        study = optuna.create_study(
            sampler=optuna.samplers.TPESampler(seed=236868),
            direction="maximize",
            pruner=_BadTrialPruner(2.0, 15),
        )
        study.optimize(
            self._objective,
            n_trials=self.n_trials,
            timeout=self.timeout,
            callbacks=[
                _StopStudyWhenConsecutivePruning(10),
                _StopStudyAfterNPruning(50),
            ],
        )

        # Parse results
        optuna_results = study.trials_dataframe()
        results = pd.DataFrame(
            {
                "date": datetime.today().strftime("%d %b %y"),
                "model": type(self.model).__name__,
                "params": [x.params for x in study.get_trials()],
                "score": optuna_results["value"],
                "worst_case": optuna_results["value"]
                - optuna_results["user_attrs_score_std"],
                "time": optuna_results["user_attrs_time"],
            }
        ).sort_values("score", ascending=False)
        self.trial_count_ = len(study.trials)

        # Warn against edge params
        warn_at_extreme(results.iloc[0]["params"], self.samples_)

        return results

    def _objective(self, trial: optuna.Trial) -> float:
        # Make a copy
        model_copy = copy.deepcopy(self.model)

        # Cross validation
        t_start = time.time()
        scores = cross_val_score(model_copy, self.x, self.y, scoring=self.scoring)
        score = sum(scores) / len(scores)
        run_time = time.time() - t_start

        # Set manual metrics
        trial.set_user_attr("time", run_time)
        trial.set_user_attr("score_std", np.std(scores))

        # Stop trial (avoid overwriting)
        if trial.number == self.n_trials:
            trial.study.stop()

        # Pruning
        if trial.should_prune():
            raise optuna.exceptions.TrialPruned

        return score


# ----------------------------------------------------------------------
# Pruning


class _BadTrialPruner(optuna.pruners.BasePruner):
    """
    Pruner to detect outlying metrics of the trials.

    Prune if the mean intermediate value is worse than the best trial value minus (or
    plus) a multiple of its intermediate value standard deviation.

    Parameters
    ----------
    std_threshold_multiplier : float
        Multiplier for the best trials intermediate value std to define the pruning
        threshold.
    n_startup_trials : int
        Pruning is disabled until the given number of trials finish in the same study.
    """

    def __init__(self, std_threshold_multiplier, n_startup_trials):
        if n_startup_trials < 0:
            raise ValueError(
                f"Number of startup trials cannot be negative but got "
                f"{n_startup_trials}."
            )

        self._std_threshold_multiplier = std_threshold_multiplier
        self._n_startup_trials = n_startup_trials

    def prune(self, study, trial):
        # Don't prune while startup trials
        if not [t for t in study.trials if t.state == optuna.trial.TrialState.COMPLETE]:
            return False
        if len(study.best_trial.intermediate_values) < self._n_startup_trials:
            return False

        # Define pruning thresholds
        best_trial_value = study.best_trial.value
        best_trial_value_std = np.std(
            list(study.best_trial.intermediate_values.values())
        )
        threshold = (
            best_trial_value - self._std_threshold_multiplier * best_trial_value_std
        )

        # Pruning
        curr_trial_mean = np.mean(list(trial.intermediate_values.values()))
        if study.direction == optuna.study.StudyDirection.MAXIMIZE:
            return curr_trial_mean < threshold
        else:
            raise RuntimeError(
                "Pruning with a threshold is arbitrary when the study direction is "
                "undefined."
            )


class _StopStudyWhenConsecutivePruning:
    """
    Optuna study callback that stops the study when trials keep being pruned.

    Parameters
    ----------
    threshold : int
        Critical threshold for consecutively pruned trials. Stops trial when achieved.
    """

    def __init__(self, threshold):
        if threshold < 0:
            raise ValueError(f"Threshold cannot be negative but got {threshold}.")

        self._threshold = threshold
        self._consecutive_pruned_count = 0

    def __call__(self, study, trial):
        """
        Stops study when consecutive prune count exceeds threshold.

        Callback function that gets called after each trial.

        Parameters
        ----------
        study : optuna.study.Study
            Study object of the target study.
        trial : optuna.trial.FrozenTrial
            FrozenTrial object of the target trial. Take a copy before modifying this
            object.
        """
        if trial.state == optuna.trial.TrialState.PRUNED:
            self._consecutive_pruned_count += 1
        else:
            self._consecutive_pruned_count = 0

        if self._consecutive_pruned_count > self._threshold:
            study.stop()


class _StopStudyAfterNPruning:
    """
    Optuna study callback that stops the study after begin N times pruned.

    Parameters
    ----------
    threshold : int
        Critical threshold for total number of pruned trials. Stops trial when achieved.
    """

    def __init__(self, threshold):
        if threshold < 0:
            raise ValueError(f"Threshold cannot be negative but got {threshold}.")

        self._threshold = threshold

    def __call__(self, study, trial):
        """
        Stops study when total prune count exceeds threshold.

        Callback function that gets called after each trial.

        Parameters
        ----------
        study : optuna.study.Study
            Study object of the target study.
        trial : optuna.trial.FrozenTrial
            FrozenTrial object of the target trial. Take a copy before modifying this
            object.
        """
        all_trials = study.get_trials(deepcopy=False)
        n_prunings = len(
            [t for t in all_trials if t.state == optuna.trial.TrialState.PRUNED]
        )

        if n_prunings > self._threshold:
            study.stop()



#  Copyright (c) 2022 by Amplo.

import json
from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from numpy.random import Generator

from amplo.automl.feature_processing.feature_processor import (
    FeatureProcessor,
    find_collinear_columns,
    get_required_columns,
    translate_features,
)
from amplo.automl.feature_processing.nop_feature_extractor import NopFeatureExtractor
from amplo.automl.feature_processing.temporal_feature_extractor import (
    TemporalFeatureExtractor,
)


@pytest.mark.usefixtures("make_rng")
class TestFunctions:
    rng: Generator

    def test_find_collinear_columns(self):
        col = np.linspace(-4, 4, 100)
        x = pd.DataFrame(
            {"a": col, "b": col, "c": col + self.rng.normal(scale=0.1, size=100)}
        )
        collinear_columns = find_collinear_columns(x, information_threshold=0.99)
        assert set(collinear_columns) == {"b", "c"}, "Collinear columns not found"

    def test_get_required_columns(self):
        feat_cols = ["inv__A", "A__mul__B"]
        assert get_required_columns(feat_cols) == ["A", "B"]

    def test_translate_features(self):
        # Accept all possible feature names
        translation = {
            "A": ["A"],
            "B__mul__C": ["B", "C"],
            "D__div__E": ["D", "E"],
            "F__x__G": ["F", "G"],
            "H__d__I": ["H", "I"],
            "sin__J": ["J"],
            "cos__K": ["K"],
            "inv__L": ["L"],
            "M__pool=sum": ["M"],
            "N__wav__cmor__pool=sum": ["N"],
        }
        assert translate_features(list(translation.keys())) == translation

        # Don't accept bad feature names
        with pytest.raises(ValueError):
            translate_features(["unknown__feature__name"])


@pytest.mark.usefixtures("make_rng")
class TestFeatureProcessor:
    rng: Generator

    @pytest.mark.parametrize("mode", ["classification", "multiclass", "regression"])
    @pytest.mark.parametrize("extraction", ["nop", "static", "temporal"])
    def test_transform(self, mode, make_data, extraction):
        data = make_data
        kwargs = {"mode": mode, "analyse_feature_sets": None}
        if extraction == "nop":
            fp = FeatureProcessor(extract_features=False, **kwargs)
        elif extraction == "static":
            fp = FeatureProcessor(extract_features=True, is_temporal=False, **kwargs)
        elif extraction == "temporal":
            index = pd.MultiIndex.from_product([[0, 1], range(len(data) // 2)])
            data.index = index
            kwargs = {**kwargs, "timeout": 1}
            fp = FeatureProcessor(extract_features=True, is_temporal=True, **kwargs)
        else:
            raise ValueError("Invalid extraction mode.")

        # Test equivalence of transform and fit_transform
        out1 = fp.fit_transform(data)
        out2 = fp.transform(data)
        assert all(out1 == out2)

        # Test transform_target
        if isinstance(fp.feature_extractor, TemporalFeatureExtractor):
            y_transformed = fp.feature_extractor._fit_data_to_window_size(
                data["target"]
            )
            y_transformed = fp.feature_extractor._pool_target(y_transformed)
        else:
            y_transformed = data["target"]
        assert np.allclose(y_transformed, fp.transform_target(data["target"]))

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_settings(self, mode, make_data):
        data = make_data
        fp = FeatureProcessor(
            mode=mode, extract_features=False, analyse_feature_sets=None
        )
        fp.fit(data)

        # Test load settings directly
        new_fp = FeatureProcessor().load_settings(fp.get_settings())
        assert fp.get_settings() == new_fp.get_settings()
        assert all(fp.transform(data) == new_fp.transform(data))

        # Test JSON serializable
        settings = json.loads(json.dumps(fp.get_settings()))
        new_fp = FeatureProcessor().load_settings(settings)
        assert fp.get_settings() == new_fp.get_settings()
        assert all(fp.transform(data) == new_fp.transform(data))

    @pytest.mark.parametrize("is_temporal", [True, False])
    def test_check_is_temporal_attribute(self, is_temporal):
        size = 100
        if is_temporal:
            index = pd.MultiIndex.from_product([[0], range(size)])
        else:
            index = pd.RangeIndex(size)
        data = pd.DataFrame(index=index)

        fp = FeatureProcessor()
        assert fp.is_temporal is None, "Attribute shouldn't be set yet."
        fp._check_is_temporal_attribute(data)
        assert fp.is_temporal == is_temporal, "Attribute is set wrong."

    def test_find_columns_of_interest(self):
        x = pd.DataFrame(
            {
                "A": [1, 4, 1],
                "B": [10, 9, 1],
                "C": [datetime.now()] * 3,
                "D": [39, 0, 0],
                "collinear": [1, 4, 1],
            }
        )
        fp = FeatureProcessor(extract_features=False)
        fp._find_columns_of_interest(x)
        fp.feature_extractor = NopFeatureExtractor()
        fp.features_ = ["A__mul__B", "inv__A"]
        assert set(fp.numeric_cols_) == set(list("ABD"))
        assert set(fp.datetime_cols_) == {"C"}
        assert set(fp.collinear_cols_) == {"collinear"}
        assert set(get_required_columns(fp.features_)) == set(list("AB"))

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    @pytest.mark.parametrize("analyse_fs", [None, "auto", "all", "gini", "shap"])
    def test_feature_sets(self, mode, analyse_fs):
        size = 100
        data = pd.DataFrame({"zeros": np.zeros(size)})
        data["target"] = pd.Series(self.rng.choice([0, 1], size))
        data["y"] = data["target"]
        fp = FeatureProcessor(
            target="target",
            mode=mode,
            extract_features=False,
            analyse_feature_sets=analyse_fs,
            selection_cutoff=0.99,
            selection_increment=0.01,
        ).fit(data)

        if analyse_fs is None:
            assert set(fp.feature_sets_) == {"take_all"}
            assert set(fp.feature_sets_["take_all"]) == {"y", "zeros"}
        else:
            # Check validity of feature sets
            for name, values in fp.feature_sets_.items():
                assert set(values) == {"y"}, "Invalid feature set."
                x_out = fp.transform(data, feature_set=name)
                assert all(
                    x_out["y"] == data["y"]
                ), "Erroneous feature set transformation."

            # Check that desired feature sets are present
            expected_fs = set()
            if analyse_fs in ("auto", "all", "gini"):
                expected_fs = {*expected_fs, "rf_increment", "rf_threshold"}
            if analyse_fs in ("auto", "all", "shap"):
                expected_fs = {*expected_fs, "shap_increment", "shap_threshold"}
            assert set(fp.feature_sets_) == expected_fs

        # Additional test for shap features:
        # When data size is >= 5000 samples, don't calculate shap features.
        if analyse_fs == "auto":
            size = 50_000
            y = pd.Series(self.rng.choice([0, 1], size))
            x = pd.DataFrame({"y": y, "zeros": np.zeros(size)})
            fp._analyse_feature_sets(x, y)
            assert "shap_increment" not in fp.feature_sets_
            assert "shap_threshold" not in fp.feature_sets_

    def test_impute_missing_columns(self):
        # Make data
        size = 100
        numeric = pd.DataFrame(
            self.rng.normal(size=(size, 5)), columns=[f"numeric_{i}" for i in range(5)]
        )
        date = pd.DataFrame(np.array([datetime.now()] * size), columns=["datetime_0"])
        collinear = numeric.rename(
            columns={col: f"collinear_{i}" for i, col in enumerate(numeric)}
        )

        # Init feature processor
        fp = FeatureProcessor(mode="classification")
        fp._is_fitted = True
        fp.feature_extractor = NopFeatureExtractor()
        fp.feature_extractor._is_fitted = True
        fp.numeric_cols_ = list(numeric.columns)
        fp.datetime_cols_ = list(date.columns)
        fp.collinear_cols_ = list(collinear.columns)
        fp.features_ = [*fp.numeric_cols_, *fp.datetime_cols_]

        # Check no imputation
        x = pd.concat([numeric, date], axis=1)
        imputed = fp._impute_missing_columns(x)
        assert set(x) == set(imputed), "Not all / too many columns are present."
        assert all(x == imputed), "Invalid imputation of values."

        # Check numeric imputation
        x = pd.concat([numeric, date], axis=1)
        imputed = fp._impute_missing_columns(x.drop("numeric_0", axis=1))
        x.loc[:, "numeric_0"] = 0
        assert set(x) == set(imputed), "Not all / too many columns are present."
        assert all(x == imputed[x.columns]), "Erroneous numeric value imputation."
